import argparse
import logging
import shutil
import sys
from os import getenv
from pathlib import Path

from buildingmotif import BuildingMOTIF
from buildingmotif.dataclasses import Library
from buildingmotif.ingresses.bacnet import BACnetNetwork

cli = argparse.ArgumentParser(
    prog="buildingmotif", description="CLI Interface for common BuildingMOTIF tasks"
)
subparsers = cli.add_subparsers(dest="subcommand")
subcommands = {}
log = logging.getLogger()
log.setLevel(logging.INFO)

ONTOLOGY_FILE_SUFFIXES = ["ttl", "n3", "ntriples", "xml"]


# borrowing some ideas from https://gist.github.com/mivade/384c2c41c3a29c637cb6c603d4197f9f
def arg(*argnames, **kwargs):
    """Helper for defining arguments on subcommands"""
    return argnames, kwargs


def subcommand(*subparser_args, parent=subparsers):
    """Decorates a function and makes it available as a subcommand"""

    def decorator(func):
        parser = parent.add_parser(func.__name__, description=func.__doc__)
        for args, kwargs in subparser_args:
            parser.add_argument(*args, **kwargs)
        parser.set_defaults(func=func)
        subcommands[func] = parser

    return decorator


def get_db_uri(args) -> str:
    """
    Fetches the db uri from args, or prints the usage
    for the corresponding subcommand
    """
    db_uri = args.db
    if db_uri is not None:
        return db_uri
    db_uri = getenv("DB_URI")
    if db_uri is not None:
        return db_uri
    print("No DB URI could be found")
    subcommands[args.func].print_help()
    sys.exit(1)


@subcommand(
    arg(
        "-d",
        "--db",
        help="Database URI of the BuildingMOTIF installation. "
        'Defaults to $DB_URI and then contents of "config.py"',
    ),
    arg(
        "--dir",
        help="Path to a local directory containing the library",
        nargs="+",
    ),
    arg(
        "-o",
        "--ont",
        help="Remote URL or local file path to an RDF ontology",
        nargs="+",
    ),
    arg(
        "-l",
        "--libraries",
        help="Filename of the libraries YAML file specifying what "
        "should be loaded into BuildingMOTIF",
        nargs="+",
        dest="library_manifest_file",
    ),
)
def load(args):
    """
    Loads libraries from (1) local directories (--dir),
    (2) local or remote ontology files (--ont)
    (3) library spec file (--libraries): the provided YML file into the
      BuildingMOTIF instance at $DB_URI.
      Use 'get_default_libraries_yml' for the format of the expected libraries.yml file
    """
    db_uri = get_db_uri(args)
    bm = BuildingMOTIF(db_uri)
    bm.setup_tables()
    for directory in args.dir or []:
        Library.load(directory=directory)
    for ont in args.ont or []:
        Library.load(ontology_graph=ont)
    for library_manifest_file in args.library_manifest_file or []:
        manifest_path = Path(library_manifest_file)
        log.info(f"Loading buildingmotif libraries listed in {manifest_path}")
        Library.load_from_libraries_yml(str(manifest_path))
    bm.session.commit()


@subcommand()
def get_default_libraries_yml(_args):
    """
    Creates a default 'libraries.default.yml' file in the current directory
    that can be edited and used with buildingmotif
    """
    default_file = (
        Path(__file__).resolve().parents[1] / "resources" / "libraries.default.yml"
    )
    shutil.copyfile(default_file, "libraries.default.yml")
    print("libraries.default.yml created in the current directory")


@subcommand(
    arg(
        "-b",
        "--bind",
        help="Address on which to bind the API server",
        default="localhost",
    ),
    arg(
        "-p", "--port", help="Listening port for the API server", type=int, default=5000
    ),
    arg(
        "-d",
        "--db",
        help="Database URI of the BuildingMOTIF installation. "
        'Defaults to $DB_URI and then contents of "config.py"',
    ),
)
def serve(args):
    """
    Serves the BuildingMOTIF API on the indicated host:port
    """
    from buildingmotif.api.app import create_app

    db_uri = get_db_uri(args)
    webapp = create_app(db_uri)
    webapp.run(host=args.host, port=args.port, threaded=False)


def app():
    args = cli.parse_args()
    if args.subcommand is None:
        cli.print_help()
    else:
        args.func(args)


@subcommand(
    arg("-o", "--output_file", help="Output file for BACnet scan", required=True),
    arg("-ip", help="ip address of BACnet network to scan", default=None),
)
def scan(args):
    """Scans a BACnet network and generates a JSON file for later processing"""
    bacnet_network = BACnetNetwork(args.ip)
    bacnet_network.dump(Path(args.output_file))


# entrypoint is actually defined in pyproject.toml; this is here for convenience/testing
if __name__ == "__main__":
    app()



import re
from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from itertools import chain
from secrets import token_hex
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, Set, Tuple, Union

import rdflib
from pyshacl.helper.path_helper import shacl_path_to_sparql_path
from rdflib import Graph, URIRef
from rdflib.collection import Collection
from rdflib.term import BNode, Node

from buildingmotif import get_building_motif
from buildingmotif.dataclasses.shape_collection import ShapeCollection
from buildingmotif.namespaces import CONSTRAINT, PARAM, RDF, SH, A, bind_prefixes
from buildingmotif.utils import (
    _gensym,
    _guarantee_unique_template_name,
    get_template_parts_from_shape,
    replace_nodes,
)

if TYPE_CHECKING:
    from buildingmotif.dataclasses import Library, Model, Template


@dataclass(frozen=True)
class GraphDiff:
    """An abstraction of a SHACL Validation Result that can produce a template
    that resolves the difference between the expected and actual graph.

    Each GraphDiff has a 'focus' that is the node in the model that the
    GraphDiff is about. If 'focus' is None, then the GraphDiff is about the
    model itself rather than a specific node
    """

    # the node that failed (shape target)
    focus: Optional[URIRef]
    # the SHACL validation result graph corresponding to this failure
    validation_result: Graph
    graph: Graph

    def __post_init__(self):
        bind_prefixes(self.graph)

    def resolve(self, lib: "Library") -> List["Template"]:
        """Produces a list of templates to resolve this GraphDiff.

        :param lib: the library to hold the templates
        :type lib: Library
        :return: templates that reconcile the GraphDiff
        :rtype: List[Template]
        """
        raise NotImplementedError

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        raise NotImplementedError

    @cached_property
    def _result_uri(self) -> Node:
        """Return the 'name' of the ValidationReport to make failed_shape/failed_component
        easier to express. We compute this by taking advantage of the fact that the validation
        result graph is actually a tree with a single root. We can find the root by finding
        all URIs which appear as subjects in the validation_result graph that do *not* appear
        as objects; this should  be exactly one URI which is the 'root' of the validation result
        graph
        """
        return next(self.validation_result.subjects(RDF.type, SH.ValidationResult))

    @cached_property
    def failed_shape(self) -> Optional[URIRef]:
        """The URI of the Shape that failed"""
        return self.validation_result.value(self._result_uri, SH.sourceShape)

    @cached_property
    def failed_component(self) -> Optional[URIRef]:
        """The Constraint Component of the Shape that failed"""
        return self.validation_result.value(
            self._result_uri, SH.sourceConstraintComponent
        )

    def __hash__(self):
        return hash(self.reason())

    def format_count_error(
        self, max_count, min_count, path, object_type: Optional[str] = None
    ) -> str:
        """Format a count error message for a given object type and path.

        :param max_count: the maximum number of objects expected
        :type max_count: int
        :param min_count: the minimum number of objects expected
        :type min_count: int
        :param object_type: the type of object expected
        :type object_type: str
        :param path: the path to the object
        :type path: str
        :return: the formatted error message
        :rtype: str
        """
        instances = f"instance(s) of {object_type} on" if object_type else "uses of"
        if min_count == max_count:
            return f"{self.focus} expected {min_count} {instances} path {path}"
        elif min_count is not None and max_count is not None:
            return f"{self.focus} expected between {min_count} and {max_count} {instances} path {path}"
        elif min_count is not None:
            return f"{self.focus} expected at least {min_count} {instances} path {path}"
        elif max_count is not None:
            return f"{self.focus} expected at most {max_count} {instances} path {path}"
        else:
            return f"{self.focus} expected {instances} path {path}"


@dataclass(frozen=True)
class OrShape(GraphDiff):
    """Represents an entity that is missing one of several possible shapes, via sh:or"""

    shapes: Tuple[URIRef]

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        return f"{self.focus} needs to match one of the following shapes: {', '.join(self.shapes)}"

    @classmethod
    def from_validation_report(cls, report: Graph) -> List["OrShape"]:
        """Construct OrShape objects from a SHACL validation report.

        :param report: the SHACL validation report
        :type report: Graph
        :return: a list of OrShape objects
        :rtype: List[OrShape]
        """
        query = """
        PREFIX sh: <http://www.w3.org/ns/shacl#>
        SELECT ?result ?focus ?shapes WHERE {
            ?result sh:sourceConstraintComponent sh:OrConstraintComponent .
            ?result sh:sourceShape/sh:or ?shapes .
            ?result sh:focusNode ?focus .
        }"""
        results = report.query(query)
        ret = []
        for result, focus, shapes in results:
            validation_report = report.cbd(result)
            ret.append(
                cls(
                    focus,
                    validation_report,
                    report,
                    tuple([s for s in Collection(report, shapes)]),
                )
            )
        return ret


@dataclass(frozen=True)
class PathClassCount(GraphDiff):
    """Represents an entity missing paths to objects of a given type:
    $this <path> <object> .
    <object> a <classname> .
    """

    path: URIRef
    minc: Optional[int]
    maxc: Optional[int]
    classname: URIRef

    @classmethod
    def from_validation_report(cls, report: Graph) -> List["PathClassCount"]:
        """Construct PathClassCount objects from a SHACL validation report.

        :param report: the SHACL validation report
        :type report: Graph
        :return: a list of PathClassCount objects
        :rtype: List[PathClassCount]
        """

        query = """
        PREFIX sh: <http://www.w3.org/ns/shacl#>
        SELECT ?focus ?path ?minc ?maxc ?classname WHERE {
            ?result sh:sourceShape/sh:qualifiedValueShape? ?shape .
            { ?result sh:sourceConstraintComponent sh:CountConstraintComponent }
            UNION
            { ?result sh:sourceConstraintComponent sh:QualifiedMinCountConstraintComponent }
            ?result sh:focusNode ?focus .
            ?shape sh:resultPath ?path .
            {
                ?shape sh:class ?classname .
                ?shape sh:minCount ?minc .
                OPTIONAL { ?shape sh:maxCount ?maxc }
            }
            UNION
            {
                ?shape sh:qualifiedValueShape [ sh:class ?classname ] .
                ?shape sh:qualifiedMinCount ?minc .
                OPTIONAL { ?shape sh:qualifiedMaxCount ?maxc }
            }
        }"""
        results = report.query(query)
        return [
            cls(
                focus,
                report,
                report,
                path,
                minc,
                maxc,
                classname,
            )
            for focus, path, minc, maxc, classname in results
        ]

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        # interpret a SHACL property path as a sparql property path
        path = shacl_path_to_sparql_path(
            self.graph, self.path, prefixes=dict(self.graph.namespaces())
        )

        classname = self.graph.qname(self.classname)
        return self.format_count_error(self.maxc, self.minc, path, classname)

    def resolve(self, lib: "Library") -> List["Template"]:
        """Produces a list of templates to resolve this GraphDiff.

        :param lib: the library to hold the templates
        :type lib: Library
        :return: templates that reconcile the GraphDiff
        :rtype: List[Template]
        """
        assert self.focus is not None
        body = Graph()
        # extract everything after the last "delimiter" character from self.classname
        name = re.split(r"[#\/]", self.classname)[-1]
        focus = re.split(r"[#\/]", self.focus)[-1]
        for _ in range(self.minc or 0):
            inst = _gensym()
            body.add((self.focus, self.path, inst))
            body.add((inst, A, self.classname))
        template_name = _guarantee_unique_template_name(lib, f"resolve{focus}{name}")
        return [lib.create_template(template_name, body)]


@dataclass(frozen=True, unsafe_hash=True)
class PathShapeCount(GraphDiff):
    """Represents an entity missing paths to objects that match a given shape.
    $this <path> <object> .
    <object> a <shapename> .
    """

    path: URIRef = field(hash=True)
    minc: Optional[int] = field(hash=True)
    maxc: Optional[int] = field(hash=True)
    shapename: URIRef = field(hash=True)
    extra_body: Optional[Graph] = field(hash=False)
    extra_deps: Optional[Tuple] = field(hash=False)

    @classmethod
    def from_validation_report(
        cls, report: Graph
    ) -> Generator["PathShapeCount", None, None]:
        """Construct PathShapeCount objects from a SHACL validation report.

        :param report: the SHACL validation report
        :type report: Graph
        :return: a list of PathShapeCount objects
        :rtype: List[PathShapeCount]
        """
        query = """
        PREFIX sh: <http://www.w3.org/ns/shacl#>
        SELECT ?focus ?path ?minc ?maxc ?shapename WHERE {
            ?result sh:sourceShape ?shape .
            ?result sh:resultPath ?path .
            { ?result sh:sourceConstraintComponent sh:CountConstraintComponent }
            UNION
            { ?result sh:sourceConstraintComponent sh:QualifiedMinCountConstraintComponent }
            ?result sh:focusNode ?focus .
            {
                ?shape sh:node ?shapename .
                ?shape sh:minCount ?minc .
                OPTIONAL { ?shape sh:maxCount ?maxc }
            }
            UNION
            {
                ?shape sh:qualifiedValueShape [ sh:node ?shapename ] .
                ?shape sh:qualifiedMinCount ?minc .
                OPTIONAL { ?shape sh:qualifiedMaxCount ?maxc }
            }

        }"""
        results = report.query(query)
        for (focus, path, minc, maxc, shapename) in results:
            extra_body, deps = get_template_parts_from_shape(shapename, report)
            yield cls(
                focus,
                report,
                report,
                path,
                minc,
                maxc,
                shapename,
                extra_body,
                tuple(deps),
            )

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        shapename = self.graph.qname(self.shapename)
        return self.format_count_error(self.maxc, self.minc, self.path, shapename)

    def resolve(self, lib: "Library") -> List["Template"]:
        """Produces a list of templates to resolve this GraphDiff."""
        assert self.focus is not None
        generated = []
        if self.extra_deps:
            for dep in self.extra_deps:
                dep["args"] = {k: str(v)[len(PARAM) :] for k, v in dep["args"].items()}
        # extract everything after the last "delimiter" character from self.shapename
        name = re.split(r"[#\/]", self.shapename)[-1]
        focus = re.split(r"[#\/]", self.focus)[-1]
        for _ in range(self.minc or 0):
            body = Graph()
            inst = PARAM["name"]
            body.add((self.focus, self.path, inst))
            body.add((inst, A, self.shapename))
            if self.extra_body:
                replace_nodes(self.extra_body, {PARAM.name: inst})
                body += self.extra_body
            template_name = _guarantee_unique_template_name(
                lib, f"resolve{focus}{name}"
            )
            templ = lib.create_template(template_name, body)
            if self.extra_deps:
                from buildingmotif.dataclasses.template import Template

                bm = get_building_motif()
                for dep in self.extra_deps:
                    dbt = bm.table_connection.get_db_template_by_name(dep["template"])
                    t = Template.load(dbt.id)
                    templ.add_dependency(t, dep["args"])
            generated.append(templ)
        return generated


@dataclass(frozen=True)
class RequiredPath(GraphDiff):
    """Represents an entity missing a required property."""

    path: URIRef
    minc: Optional[int]
    maxc: Optional[int]

    @classmethod
    def from_validation_report(cls, report: Graph) -> List["RequiredPath"]:
        """Construct RequiredPath objects from a SHACL validation report.

        :param report: the SHACL validation report
        :type report: Graph
        :return: a list of RequiredPath objects
        :rtype: List[RequiredPath]
        """
        query = """
        PREFIX sh: <http://www.w3.org/ns/shacl#>
        SELECT ?focus ?path ?minc ?maxc WHERE {
            ?result sh:sourceShape ?shape .
            ?result sh:resultPath ?path .
            { ?result sh:sourceConstraintComponent sh:CountConstraintComponent }
            UNION
            { ?result sh:sourceConstraintComponent sh:QualifiedMinCountConstraintComponent }
            ?result sh:focusNode ?focus .
            {
                ?shape sh:minCount ?minc .
                OPTIONAL { ?shape sh:maxCount ?maxc }
            } UNION {
                ?shape sh:qualifiedMinCount ?minc .
                OPTIONAL { ?shape sh:qualifiedMaxCount ?maxc }
            }
        }"""
        results = report.query(query)
        return [
            cls(
                focus,
                report,
                report,
                path,
                minc,
                maxc,
            )
            for focus, path, minc, maxc in results
        ]

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        path = shacl_path_to_sparql_path(
            self.graph, self.path, prefixes=dict(self.graph.namespaces())
        )
        return self.format_count_error(self.maxc, self.minc, path)

    def resolve(self, lib: "Library") -> List["Template"]:
        """Produces a list of templates to resolve this GraphDiff.

        :param lib: the library to hold the templates
        :type lib: Library
        :return: templates that reconcile the GraphDiff
        :rtype: List[Template]
        """
        assert self.focus is not None
        body = Graph()
        # extract everything after the last "delimiter" character from self.shapename
        name = re.split(r"[#\/]", self.path)[-1]
        focus = re.split(r"[#\/]", self.focus)[-1]
        for _ in range(self.minc or 0):
            inst = _gensym()
            body.add((self.focus, self.path, inst))
        template_name = _guarantee_unique_template_name(lib, f"resolve{focus}{name}")
        return [lib.create_template(template_name, body)]


@dataclass(frozen=True)
class RequiredClass(GraphDiff):
    """Represents an entity that should be an instance of the class."""

    classname: URIRef

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        value_node = self.validation_result.value(self._result_uri, SH.value)
        return f"{value_node} on {self.focus} needs to be a {self.classname}"

    def resolve(self, lib: "Library") -> List["Template"]:
        """Produces a list of templates to resolve this GraphDiff.

        :param lib: the library to hold the templates
        :type lib: Library
        :return: templates that reconcile the GraphDiff
        :rtype: List[Template]
        """
        assert self.focus is not None
        body = Graph()
        name = re.split(r"[#\/]", self.classname)[-1]
        body.add((self.focus, A, self.classname))
        template_name = _guarantee_unique_template_name(lib, f"resolveSelf{name}")
        return [lib.create_template(template_name, body)]


@dataclass(frozen=True)
class GraphClassCardinality(GraphDiff):
    """Represents a graph that is missing an expected number of instances of
    the given class.
    """

    classname: URIRef
    expectedCount: int

    def reason(self) -> str:
        """Human-readable explanation of this GraphDiff."""
        return f"Graph did not have {self.expectedCount} instances of {self.classname}"

    def resolve(self, lib: "Library") -> List["Template"]:
        """Produces a list of templates to resolve this GraphDiff.

        :param lib: the library to hold the templates
        :type lib: Library
        :return: templates that reconcile the GraphDiff
        :rtype: List[Template]
        """
        templs = []
        name = re.split(r"[#\/]", self.classname)[-1]
        for _ in range(self.expectedCount):
            template_body = Graph()
            template_body.add((PARAM["name"], A, self.classname))
            template_name = _guarantee_unique_template_name(lib, f"resolveAdd{name}")
            templs.append(lib.create_template(template_name, template_body))
        return templs


@dataclass
class ValidationContext:
    """Holds the necessary information for processing the results of SHACL
    validation.
    """

    shape_collections: List[ShapeCollection]
    # the shapes graph that was used to validate the model
    # This will be skolemized!
    shapes_graph: Graph
    valid: bool
    report: rdflib.Graph
    report_string: str
    model: "Model"

    @cached_property
    def diffset(self) -> Dict[Optional[URIRef], Set[GraphDiff]]:
        """The unordered set of GraphDiffs produced from interpreting the input
        SHACL validation report.
        """
        return self._report_to_diffset()

    def as_templates(self) -> List["Template"]:
        """Produces the set of templates that reconcile the GraphDiffs from the
        SHACL validation report.

        :return: reconciling templates
        :rtype: List[Template]
        """
        return diffset_to_templates(self.diffset)

    def get_broken_entities(self) -> Set[URIRef]:
        """Get the set of entities that are broken in the model.

        :return: set of entities that are broken
        :rtype: Set[URIRef]
        """
        return {diff or "Model" for diff in self.diffset}

    def get_diffs_for_entity(self, entity: URIRef) -> Set[GraphDiff]:
        """Get the set of diffs for a specific entity.

        :param entity: the entity to get diffs for
        :type entity: URIRef
        :return: set of diffs for the entity
        :rtype: Set[GraphDiff]
        """
        return self.diffset.get(entity, set())

    def get_reasons_with_severity(
        self, severity: Union[URIRef, str]
    ) -> Dict[Optional[URIRef], Set[GraphDiff]]:
        """
        Like diffset, but only includes ValidationResults with the given severity.
        Permitted values are:
        - SH.Violation or "Violation" for violations
        - SH.Warning or "Warning" for warnings
        - SH.Info or "Info" for info

        :param severity: the severity to filter by
        :type severity: Union[URIRef|str]
        :return: a dictionary of focus nodes to the reasons with the given severity
        :rtype: Dict[Optional[URIRef], Set[GraphDiff]]
        """

        if not isinstance(severity, URIRef):
            severity = SH[severity]

        # check if the severity is a valid SHACL severity
        if severity not in {SH.Violation, SH.Warning, SH.Info}:
            raise ValueError(
                f"Invalid severity: {severity}. Must be one of SH.Violation, SH.Warning, or SH.Info"
            )

        # for each value in the diffset, filter out the diffs that don't have the given severity
        # in the diffset.graph
        return {
            focus: {
                diff
                for diff in diffs
                if diff.validation_result.value(diff._result_uri, SH.resultSeverity)
                == severity
            }
            for focus, diffs in self.diffset.items()
        }

    def _report_to_diffset(self) -> Dict[Optional[URIRef], Set[GraphDiff]]:
        """Interpret a SHACL validation report and say what is missing.

        :return: a set of GraphDiffs that each abstract a SHACL shape violation
        :rtype: Set[GraphDiff]
        """
        classpath = SH["class"] | (SH.qualifiedValueShape / SH["class"])  # type: ignore
        shapepath = SH["node"] | (SH.qualifiedValueShape / SH["node"])  # type: ignore
        # TODO: for future use
        # proppath = SH["property"] | (SH.qualifiedValueShape / SH["property"])  # type: ignore

        g = self.report + self.shapes_graph
        diffs: Dict[Optional[URIRef], Set[GraphDiff]] = defaultdict(set)

        for result in g.objects(predicate=SH.result):
            # check if the failure is due to our count constraint component
            focus = g.value(result, SH.focusNode)
            # get the subgraph corresponding to this ValidationReport -- see
            # https://www.w3.org/TR/shacl/#results-validation-result for details
            # on the structure and expected properties
            validation_report = g.cbd(result)
            if (
                g.value(result, SH.sourceConstraintComponent)
                == CONSTRAINT.countConstraintComponent
            ):
                expected_count = g.value(
                    result, SH.sourceShape / CONSTRAINT.exactCount  # type: ignore
                )
                of_class = g.value(result, SH.sourceShape / CONSTRAINT["class"])  # type: ignore
                # here, our 'self.focus' is the graph itself, which we don't want to have bound
                # to the templates during evaluation (for this specific kind of diff).
                # For this reason we override focus to be None
                diffs[None].add(
                    GraphClassCardinality(
                        None, validation_report, g, of_class, int(expected_count)
                    )
                )
            elif (
                g.value(result, SH.sourceConstraintComponent)
                == SH.ClassConstraintComponent
            ):
                requiring_shape = g.value(result, SH.sourceShape)
                expected_class = g.value(requiring_shape, SH["class"])
                if expected_class is None or isinstance(expected_class, BNode):
                    continue
                diffs[focus].add(
                    RequiredClass(focus, validation_report, g, expected_class)
                )
            elif (
                g.value(result, SH.sourceConstraintComponent)
                == SH.NodeConstraintComponent
            ):
                # TODO: handle node constraint components
                pass
            # check if property shape
            elif g.value(result, SH.resultPath):
                path = g.value(result, SH.resultPath)
                min_count = g.value(
                    result, SH.sourceShape / (SH.minCount | SH.qualifiedMinCount)  # type: ignore
                )
                max_count = g.value(
                    result, SH.sourceShape / (SH.maxCount | SH.qualifiedMaxCount)  # type: ignore
                )
                classname = g.value(
                    result,
                    SH.sourceShape / classpath,
                )

                # TODO: finish this for some shapes
                # shapes_of_object = g.value(result, SH.sourceShape / SH.qualifiedValueShape)
                # for soo in shapes_of_object:
                #     soo_graph = g.cbd(soo)
                # handle properties (on qualifiedValueShapes?)
                # extra = g.value(result, SH.sourceShape / proppath)  # type: ignore

                if focus and (min_count or max_count) and classname:
                    diffs[focus].add(
                        PathClassCount(
                            focus,
                            validation_report,
                            g,
                            path,
                            int(min_count) if min_count else None,
                            int(max_count) if max_count else None,
                            classname,
                        )
                    )
                    continue
                shapename = g.value(result, SH.sourceShape / shapepath)  # type: ignore
                if focus and (min_count or max_count) and shapename:
                    extra_body, deps = get_template_parts_from_shape(shapename, g)
                    diffs[focus].add(
                        PathShapeCount(
                            focus,
                            validation_report,
                            g,
                            path,
                            int(min_count) if min_count else None,
                            int(max_count) if max_count else None,
                            shapename,
                            extra_body,
                            tuple(deps),
                        )
                    )
                    continue
                if focus and (min_count or max_count):
                    diffs[focus].add(
                        RequiredPath(
                            focus,
                            validation_report,
                            g,
                            path,
                            int(min_count) if min_count else None,
                            int(max_count) if max_count else None,
                        )
                    )

        # TODO: this is still kind of broken...ideally we would actually interpret the shapes
        # inside the or clause
        candidates = OrShape.from_validation_report(g)
        for c in candidates:
            diffs[c.focus].add(c)
        return diffs


def diffset_to_templates(
    grouped_diffset: Dict[Optional[URIRef], Set[GraphDiff]]
) -> List["Template"]:
    """Combine GraphDiff by focus node to generate a list of templates that
    reconcile what is "wrong" with the Graph with respect to the GraphDiffs.

    :param diffset: a set of diffs produced by `_report_to_diffset`
    :type diffset: Set[GraphDiff]
    :return: list of templates that should resolve the SHACL violations when
        populated
    :rtype: List[Template]
    """
    from buildingmotif.dataclasses import Library, Template

    lib = Library.create(f"resolve_{token_hex(4)}")

    templates = []
    # now merge all tempaltes together for each focus node
    for focus, diffset in grouped_diffset.items():
        if focus is None:
            for diff in diffset:
                templates.extend(diff.resolve(lib))
            continue

        templ_lists = (diff.resolve(lib) for diff in diffset)
        templs: List[Template] = list(filter(None, chain.from_iterable(templ_lists)))
        if len(templs) <= 1:
            templates.extend(templs)
            continue
        base = templs[0]
        # treat all the other templates as dependencies of the first one.
        # This allows us to do a "join" with inline_dependencies() which
        # will ensure that there are no unintended overlaps in the choice
        # of parameter name
        for templ in templs[1:]:
            # if there is a 'name' in the parameter list, join on that name.
            # otherwise, just append the body
            # (we don't need to use use to_inline() to ensure uniqueness of parameters
            # because all params are created with _gensym() which ensures uniqueness)
            if "name" in templ.parameters:
                base.add_dependency(templ, {"name": "name"})
            else:
                base.body += templ.body
        unified = base.inline_dependencies()
        # only try to evaluate if there are parameters, else this will fail.
        # We may not have parameters if the GraphDiffs have all the information
        # they need to patch the graph and don't need user input
        if len(unified.parameters) > 0:
            unified_evaluated = unified.evaluate({"name": focus})
        else:
            unified_evaluated = unified
        assert isinstance(unified_evaluated, Template)
        templates.append(unified_evaluated)
    return templates



import warnings
from functools import lru_cache
from inspect import Parameter, signature
from typing import Any, Type, Union, get_type_hints

from rdflib import URIRef
from typing_extensions import TypedDict

from buildingmotif.label_parsing import combinators  # noqa
from buildingmotif.label_parsing.parser import Parser
from buildingmotif.label_parsing.tokens import Token

ParserDict = TypedDict("ParserDict", {"parser": str, "args": dict})


def serialize(parser: Parser) -> ParserDict:
    """Serialize a parser into a TypedDict.

    :param parser: a parser
    :type parser: Parser
    :return: a dict ready for JSON serialization
    :rtype: ParserDict
    """
    parser_dict: ParserDict = {
        "parser": parser.__class__.__name__,
        "args": _serialize(parser.__args__),  # type: ignore
    }

    return parser_dict


def _serialize(collection: Union[dict, list, tuple]) -> Union[dict, list]:
    """Serialize a collection and the contents of that collection into a dict or list

    :param collection: collection to serialize
    :type collection: Union[dict, list, tuple]
    :return: serialized collection
    :rtype: Union[dict, list]
    """

    def _serialize_item(item):
        if isinstance(item, URIRef):
            return str(item)
        elif isinstance(item, str) or isinstance(item, int) or isinstance(item, float):
            return item
        elif item in [None, True, False]:
            return item
        elif isinstance(item, Parser):
            return serialize(item)
        if isinstance(item, type) and issubclass(item, Token):
            return {"token": item.__name__}
        if isinstance(item, Token):
            return {"token": item.__class__.__name__, "value": str(item.value)}
        if isinstance(item, list) or isinstance(item, dict) or isinstance(item, tuple):
            return _serialize(item)
        warnings.warn(
            f"Serialization does not exist for object of class {item.__class__} this may cause JSON serialization to fail"
        )
        return item

    serialized_collection: Union[dict, list]
    if isinstance(collection, dict):
        serialized_collection = {}
        for key, item in collection.items():
            serialized_collection[key] = _serialize_item(item)
    else:
        serialized_collection = list(map(_serialize_item, collection))
    return serialized_collection


@lru_cache
def _get_parser_by_name(parser_name: str) -> Type[Parser]:
    """Get parser class by name of parser.

    :param parser_name: name of parser
    :type parser_name: str
    :return: parser class
    :rtype: Type[Parser]"""
    parsers = Parser.__subclasses__()
    for parser in parsers:
        if parser.__name__ == parser_name:
            return parser
    raise NameError(f'Parser of type "{parser_name}" does not exist')


@lru_cache
def _get_token_by_name(token_name: str) -> Type[Token]:
    """Get token class by name of token

    :param token_name: name of token
    :type token_name: str
    :return: token class
    :rtype: Type[Token]"""
    tokens = Token.__subclasses__()
    for token in tokens:
        if token.__name__ == token_name:
            return token
    raise NameError(f'Token of type "{token_name}" does not exist')


def _construct_class(cls: Type[Parser], args_dict: dict) -> Parser:
    """Construct class from type and arguments

    :param cls: type of class to construct
    :type cls: Type
    :param args: arguments
    :type args: dict
    :return: Instance of class
    :rtype: Any"""
    parameters = signature(cls.__init__).parameters
    args = []
    kwargs = {}
    for name, param in parameters.items():
        if name not in args_dict:
            continue
        kind = param.kind
        value = args_dict[name]
        if kind == Parameter.POSITIONAL_ONLY:
            args.append(value)
        elif kind in [Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY]:
            kwargs[name] = value
        elif kind == Parameter.VAR_POSITIONAL:
            args = [*args, *value]
        elif kind == Parameter.VAR_KEYWORD:
            kwargs.update(value)

    return cls(*args, **kwargs)


def deserialize(parser_dict: Union[ParserDict, dict]) -> Parser:
    """Deserialize a parser from a TypedDict.

    :param parser_dict: dict containing serialized parser
    :type parser_dict: ParserDict
    :return: deserialized parser
    :rtype: Parser
    """
    if not _parser_like(parser_dict):
        raise TypeError("parser_dict is incorrectly formed")
    args_dict = dict(
        (arg, _deserialize(item)) for arg, item in parser_dict["args"].items()
    )
    parser = _get_parser_by_name(parser_dict["parser"])
    return _construct_class(parser, args_dict)


@lru_cache
def _get_token_value_type(token: Type[Token]) -> Type:
    """Get the type of the token's value argument.

    :param token: token to inspect
    :type token: Token
    :return: type of value argument
    :rtype: Type"""
    return get_type_hints(token.__init__)["value"]


def _deserialize_token(token_dict: dict) -> Union[Token, Type[Token]]:
    """Deserialize token dict

    :param token_dict: dict containing token
    :type token_dict: dict
    :return: deserialized token
    :rtype: Token"""
    token = _get_token_by_name(token_dict["token"])
    if "value" in token_dict:
        value_type = _get_token_value_type(token)  # type: ignore
        return token(value_type(token_dict["value"]))
    return token


def _deserialize(item: Any) -> Any:
    """Deserialize an item

    :param item: item to deserialize
    :type item: Any
    :return: deserialized item
    :rtype: Any
    """
    if isinstance(item, str) or isinstance(item, int) or isinstance(item, float):
        return item
    elif item in [None, True, False]:
        return item
    elif isinstance(item, dict):
        if _parser_like(item):
            return deserialize(item)
        elif _token_like(item):
            return _deserialize_token(item)
        else:
            return dict((key, _deserialize(value)) for key, value in item.items())
    elif isinstance(item, list):
        return list(map(_deserialize, item))
    warnings.warn(
        f"Serialized object contained unexpected type: {item.__class__.__name__}"
    )
    return item


def _parser_like(item: Union[dict, ParserDict]) -> bool:
    """Does a dict look like a serialized parser

    :param item: dict to inspect
    :type item: dict
    :return: does the dict look like a serialized parser
    :rtype: bool
    """
    if len(item) != 2:
        return False
    if "args" not in item:
        return False
    if "parser" not in item:
        return False
    return True


def _token_like(item: dict) -> bool:
    """Does the dict look like a serialized token

    :param item: dict to inspect
    :type item: dict
    :return: does the dict look like a serialized token
    :rtype: bool
    """
    if len(item) > 2:
        return False
    if "token" not in item:
        return False
    return True



import logging
import random
import string
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

import rdflib
from pyshacl.helper.path_helper import shacl_path_to_sparql_path
from pyshacl.shapes_graph import ShapesGraph
from rdflib import RDF, RDFS, Graph, URIRef
from rdflib.paths import ZeroOrMore, ZeroOrOne
from rdflib.term import Node

from buildingmotif import get_building_motif
from buildingmotif.namespaces import BMOTIF, OWL, SH
from buildingmotif.utils import Triple, copy_graph, get_template_parts_from_shape

if TYPE_CHECKING:
    from buildingmotif import BuildingMOTIF
    from buildingmotif.dataclasses import Library

ONTOLOGY_FILE = (
    Path(__file__).resolve().parents[1] / "resources" / "building_motif_ontology.ttl"
)
ontology = rdflib.Graph().parse(ONTOLOGY_FILE)


@dataclass
class ShapeCollection:
    """This class mirrors :py:class:`database.tables.DBShapeCollection`."""

    _id: int
    graph: rdflib.Graph
    _bm: "BuildingMOTIF"

    @classmethod
    def create(cls) -> "ShapeCollection":
        """Create a new ShapeCollection.

        :return: new ShapeCollection
        :rtype: ShapeCollection
        """
        bm = get_building_motif()
        db_shape_collection = bm.table_connection.create_db_shape_collection()
        graph = bm.graph_connection.create_graph(
            db_shape_collection.graph_id, rdflib.Graph()
        )

        return cls(_id=db_shape_collection.id, graph=graph, _bm=bm)

    @classmethod
    def load(cls, id: int) -> "ShapeCollection":
        """Get ShapeCollection from database by id.

        :param id: ShapeCollection id
        :type id: int
        :return: ShapeCollection
        :rtype: ShapeCollection
        """
        bm = get_building_motif()
        db_shape_collection = bm.table_connection.get_db_shape_collection(id)
        graph = bm.graph_connection.get_graph(db_shape_collection.graph_id)

        return cls(_id=db_shape_collection.id, graph=graph, _bm=bm)

    @property
    def id(self) -> Optional[int]:
        return self._id

    @id.setter
    def id(self, new_id):
        raise AttributeError("Cannot modify db id")

    @property
    def graph_name(self) -> Optional[URIRef]:
        """
        Returns the name of the graph (subject of "a owl:Ontology")
        if one exists
        """
        # will be None if this is not found
        return self.graph.value(predicate=RDF.type, object=OWL.Ontology)  # type: ignore

    def add_triples(self, *triples: Triple) -> None:
        """Add the given triples to the graph.

        :param triples: a sequence of triples to add to the graph
        :type triples: Triple
        """
        for triple in triples:
            self.graph.add(triple)

    def add_graph(self, graph: rdflib.Graph) -> None:
        """Add the given graph to the ShapeCollection.

        :param graph: the graph to add to the ShapeCollection
        :type graph: rdflib.Graph
        """
        self.graph += graph

    def _cbd(self, shape_name, self_contained=True):
        """Retrieves the Concise Bounded Description (CBD) of the shape."""
        cbd = self.graph.cbd(shape_name)
        # if computing self-contained, do the fixed-point computation produced by unioning
        # the CBDs of all nodes in the current CBD until the graph does not change
        changed = True
        while self_contained and changed:
            new_g = rdflib.Graph()
            for node in cbd.all_nodes():
                new_g += self.graph.cbd(node)
            new_cbd = new_g + cbd
            changed = len(new_cbd) > cbd
            cbd = new_cbd
        return cbd

    def resolve_imports(
        self, recursive_limit: int = -1, error_on_missing_imports: bool = True
    ) -> "ShapeCollection":
        """Resolves `owl:imports` to as many levels as requested.

        By default, all `owl:imports` are recursively resolved. This limit can
        be changed to 0 to suppress resolving imports, or to 1..n to handle
        recursion up to that limit.

        :param recursive_limit: how many levels of `owl:imports` to resolve,
            defaults to -1 (all)
        :type recursive_limit: int, optional
        :param error_on_missing_imports: if True, raises an error if any of the dependency
            ontologies are missing (i.e. they need to be loaded into BuildingMOTIF), defaults
            to True
        :type error_on_missing_imports: bool, optional
        :return: a new ShapeCollection with the types resolved
        :rtype: ShapeCollection
        """
        resolved_namespaces: Set[rdflib.URIRef] = set()
        resolved = _resolve_imports(
            self.graph,
            recursive_limit,
            resolved_namespaces,
            error_on_missing_imports=error_on_missing_imports,
        )
        new_sc = ShapeCollection.create()
        new_sc.add_graph(resolved)
        return new_sc

    @classmethod
    def _get_subclasses_of_definition_type(
        cls, definition_type: URIRef
    ) -> List[URIRef]:
        """Get all the definition types in the ontology that are subclasses in
        the given definition types.

        :param definition_type: the given definition type
        :type definition_type: URIRef
        :return: list of included definition types
        :rtype: List[URIRef]
        """
        children = ontology.subjects(RDFS.subClassOf, definition_type)

        results = [definition_type]
        for child in children:
            results += cls._get_subclasses_of_definition_type(child)

        return results

    @classmethod
    def _get_included_domains(cls, domain: URIRef) -> List[URIRef]:
        """Get all the domains in the ontology that are included in the given
        domains.

        :param domain: the given domain
        :type domain: URIRef
        :return: list of included domains
        :rtype: List[URIRef]
        """
        children = ontology.subjects(BMOTIF.includes, domain)

        results = [domain]
        for child in children:
            results += cls._get_included_domains(child)

        return results

    def infer_templates(self, library: "Library") -> None:
        """Infer templates from the graph in this ShapeCollection and add them to the given library.

        :param library: The library to add inferred templates to
        :type library: Library
        """
        # we need to do the Library import here to avoid circular imports
        from buildingmotif.dataclasses.library import Library

        imports_closure = copy_graph(self.graph)
        for dependency in self.graph.objects(predicate=rdflib.OWL.imports):
            try:
                lib = Library.load(name=str(dependency))
                imports_closure += lib.get_shape_collection().graph
            except Exception as e:
                logging.warning(
                    f"An ontology could not resolve a dependency on {dependency} ({e}). Check this is loaded into BuildingMOTIF"
                )
                continue

        class_candidates = set(self.graph.subjects(rdflib.RDF.type, rdflib.OWL.Class))
        shape_candidates = set(
            self.graph.subjects(rdflib.RDF.type, rdflib.SH.NodeShape)
        )
        candidates = class_candidates.intersection(shape_candidates)

        template_id_lookup: Dict[str, int] = {}
        dependency_cache: Dict[int, List[Dict[Any, Any]]] = {}

        for candidate in candidates:
            assert isinstance(candidate, rdflib.URIRef)
            partial_body, deps = get_template_parts_from_shape(
                candidate, imports_closure
            )
            templ = library.create_template(str(candidate), partial_body)
            dependency_cache[templ.id] = deps
            template_id_lookup[str(candidate)] = templ.id

        library._resolve_template_dependencies(template_id_lookup, dependency_cache)

    def get_shapes_of_definition_type(
        self, definition_type: URIRef, include_labels=False
    ) -> Union[List[URIRef], List[Tuple[URIRef, str]]]:
        """Get subjects present in shape of the definition type.

        :param definition_type: desired definition type
        :type definition_type: URIRef
        :return: subjects
        :rtype: List[URIRef]
        """
        definition_types = self._get_subclasses_of_definition_type(definition_type)

        results = []
        for definition_type in definition_types:
            instances = self.graph.subjects(RDF.type, definition_type)
            if include_labels:
                results += [
                    (shape, self.graph.value(shape, RDFS.label)) for shape in instances
                ]
            else:
                results += instances

        return results

    def get_shapes_of_domain(self, domain: URIRef) -> List[URIRef]:
        """Get subjects present in shape of domain type.

        :param domain: desired domain
        :type domain: URIRef
        :return: subjects
        :rtype: List[URIRef]
        """
        included_domains = self._get_included_domains(domain)

        results = []
        for domain in included_domains:
            results += self.graph.subjects(RDF.type, domain)

        return results

    def get_shapes_about_class(
        self, rdf_type: URIRef, contexts: Optional[List["ShapeCollection"]] = None
    ) -> List[URIRef]:
        """Returns a list of shapes that either target the given class (or
        superclasses of it), or otherwise only apply to URIs of the given type.

        :param rdf_type: an OWL class
        :type rdf_type: URIRef
        :param contexts: list of ShapeCollections that help determine the class
            structure
        :type contexts: List["ShapeCollection"], optional
        :return: a list of shapes in this ShapeCollection that concern that
            class
        :rtype: List[URIRef]
        """
        # merge the contexts together w/ our graph if they are provided, else
        # just use the existing shape collection graph
        if contexts is not None:
            context = sum(map(lambda x: x.graph, contexts), rdflib.Graph())
            graph = self.graph + context
        else:
            graph = self.graph
        rows = graph.query(
            f"""
            PREFIX sh: <http://www.w3.org/ns/shacl#>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            SELECT ?shape WHERE {{
            ?shape a sh:NodeShape .
            {rdf_type.n3()} rdfs:subClassOf* ?class .
            {{ ?shape sh:targetClass ?class }}
            UNION
            {{ ?shape sh:class ?class }}
        }}"""
        )
        return [row[0] for row in rows]  # type: ignore

    def shape_to_query(self, shape: URIRef) -> str:
        """
        This method takes a URI representing a SHACL shape as an argument and returns
        a SPARQL query selecting the information which would be used to satisfy that
        SHACL shape. This uses the following rules:
        - `<shape> sh:targetClass <class>` -> `?target rdf:type/rdfs:subClassOf* <class>`
        - `<shape> sh:property [ sh:path <path>; sh:class <class>; sh:name <name> ]` ->
            ?target <path> ?name . ?name rdf:type/rdfs:subClassOf* <class>
        - `<shape> sh:property [ sh:path <path>; sh:hasValue <value>]` ->
            ?target <path> <value>
        """
        clauses, project = _shape_to_where(self.graph, shape, "?target")
        preamble = """PREFIX sh: <http://www.w3.org/ns/shacl#>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        """
        return f"{preamble} SELECT DISTINCT {' '.join(project)} WHERE {{\n{clauses}\n}}"


def _is_list(graph: Graph, node: Node):
    return (node, RDF.first, None) in graph


def _target_to_sparql(graph: Graph, nodeshape: Node, root_var: str = "?target") -> str:
    """
    Takes the nodeshape and returns the SPARQL query that would be used to
    find the target nodes of that nodeshape. This is a helper function for
    _shape_to_where
    Handles:
    - targetClass
    - targetSubjectsOf
    - targetObjectsOf
    - targetNode

    If there is more than one of these clauses on the nodeshape, they are
    combined with a UNION.

    Returns the string of the query.
    """

    # get all the clauses for the targetClass
    targetClasses = graph.objects(nodeshape, SH.targetClass)
    tc_clauses = [
        f"{root_var} rdf:type/rdfs:subClassOf* {tc.n3()} .\n" for tc in targetClasses  # type: ignore
    ]
    # get all the clauses for the targetSubjectsOf
    targetSubjectsOf = graph.objects(nodeshape, SH.targetSubjectsOf)
    tso_clauses = [
        f"{root_var} {tso.n3()} ?ignore .\n" for tso in targetSubjectsOf  # type: ignore
    ]
    # get all the clauses for the targetObjectsOf
    targetObjectsOf = graph.objects(nodeshape, SH.targetObjectsOf)
    too_clauses = [
        f"?ignore {too.n3()} {root_var} .\n" for too in targetObjectsOf  # type: ignore
    ]

    # get all the clauses for the targetNode
    targetNode = list(graph.objects(nodeshape, SH.targetNode))
    tn_clauses = [
        f"BIND({tn.n3()} AS {root_var}) .\n" for tn in targetNode  # type: ignore
    ]

    # combine all the clauses with a UNION
    all_clauses = tc_clauses + tso_clauses + too_clauses + tn_clauses
    return " UNION ".join(f"{{ {clause} }}" for clause in all_clauses)


def _clauses_on_nodeshape(
    graph: Graph, nodeshape: Node, root_variable: str = "?target"
) -> str:
    """handles the constraint components on a node shape (other than targetClass, targetSubjectsOf, targetObjectsOf, targetNode).
    Builds up the SPARQL query for the given node shape, starting with the given root variable.
    """
    clauses = []
    # handle sh:class
    for class_constraint in graph.objects(nodeshape, SH["class"]):
        clauses.append(
            f"{root_variable} rdf:type/rdfs:subClassOf* {class_constraint.n3()} .\n"
        )

    return " ".join(clauses)


def _shape_to_where(
    graph: Graph, shape: URIRef, root_var: str = "?target"
) -> Tuple[str, List[str]]:
    # we will build the query as a string
    clauses: str = ""
    # build up the SELECT clause as a set of vars
    project: Set[str] = {root_var}

    # local state for generating unique variable names
    prefix = "".join(random.choice(string.ascii_lowercase) for _ in range(2))
    variable_counter = 0

    def get_varname(shape):
        """
        Uses graph.value(shape, SH.name | RDFS.label) to get a name if it exists,
        converting to a string if it is a Literal. Otherwise, generates a new unique
        variable name.
        """
        name = graph.value(shape, SH.name | RDFS.label)
        if isinstance(name, rdflib.Literal):
            name = str(name)
        if name:
            return name.replace(" ", "_")

        # generate symbol
        nonlocal variable_counter
        varname = f"{prefix}{variable_counter}"
        variable_counter += 1
        return varname

    shape_graph = ShapesGraph(graph)

    # get all the target clauses
    clauses += _target_to_sparql(graph, shape, root_var)

    clauses += _clauses_on_nodeshape(graph, shape, root_var)

    # find all of the non-qualified property shapes. All of these will use the same variable
    # for all uses of the same sh:path value
    pshapes_by_path: Dict[Node, List[Node]] = defaultdict(list)
    qualified_pshapes: Set[Node] = set()
    for pshape in graph.objects(shape, SH.property):
        path = shacl_path_to_sparql_path(shape_graph, graph.value(pshape, SH.path))
        if not graph.value(pshape, SH.qualifiedValueShape):
            pshapes_by_path[path].append(pshape)  # type: ignore
        else:
            qualified_pshapes.add(pshape)

    # look at pshapes implicitly defined by sh:path
    for pshape in graph.subjects(predicate=SH.path):
        if (
            pshape == shape
        ):  # skip the input 'shape', otherwise this will infinitely recurse
            continue
        path = shacl_path_to_sparql_path(shape_graph, graph.value(pshape, SH.path))
        if not graph.value(pshape, SH.qualifiedValueShape):
            pshapes_by_path[path].append(pshape)  # type: ignore
        else:
            qualified_pshapes.add(pshape)

    for dep_shape in graph.objects(shape, SH.node):
        dep_clause, dep_project = _shape_to_where(graph, dep_shape, root_var)
        clauses += dep_clause
        project.update(dep_project)

    for or_clause in graph.objects(shape, SH["or"]):
        items = list(graph.objects(or_clause, (RDF.rest * ZeroOrMore) / RDF.first))  # type: ignore
        or_parts = []
        for item in items:
            or_body, or_project = _shape_to_where(graph, item, root_var)
            or_parts.append(or_body)
            project.update(or_project)
        clauses += " UNION ".join(f"{{ {or_body} }}" for or_body in or_parts)

    # 'pshapes_by_path' maps a path to all of the property shapes that use that path on the target
    # assign a unique variable for each sh:path w/o a qualified shape
    pshape_vars: Dict[Node, str] = {}
    for pshape_list in pshapes_by_path.values():
        # get name if it exists, otherwise generate a new one
        pshape_name = get_varname(pshape_list[0])
        varname = f"?{pshape_name}"
        for pshape in pshape_list:
            pshape_vars[pshape] = varname

    for pshape in graph.objects(shape, SH.property):
        # get the varname if we've already assigned one for this pshape above,
        # or generate a new one. When generating a name, use the SH.name|RDFS.label field
        # in the PropertyShape or generate a unique one
        name = pshape_vars.get(
            pshape,
            f"?{get_varname(pshape)}".replace(" ", "_"),
        )
        path = shacl_path_to_sparql_path(graph, graph.value(pshape, SH.path))
        qMinCount = graph.value(pshape, SH.qualifiedMinCount) or 0

        pclass = graph.value(
            pshape, (SH["qualifiedValueShape"] * ZeroOrOne / SH["class"])  # type: ignore
        )
        if pclass:
            clause = f"{root_var} {path} {name} .\n {name} rdf:type/rdfs:subClassOf* {pclass.n3()} .\n"
            if qMinCount == 0:
                clause = f"OPTIONAL {{ {clause} }} .\n"
            clauses += clause
            project.add(name)

        pnode = graph.value(
            pshape, (SH["qualifiedValueShape"] * ZeroOrOne / SH["node"])  # type: ignore
        )
        if pnode:
            node_clauses, node_project = _shape_to_where(graph, pnode, root_var)
            clause = f"{root_var} {path} {name} .\n"
            clause += node_clauses.replace(root_var, name)
            if qMinCount == 0:
                clause = f"OPTIONAL {{ {clause} }}"
            clauses += clause
            project.update({p.replace(root_var, name) for p in node_project})

        or_values = graph.value(
            pshape, (SH["qualifiedValueShape"] * ZeroOrOne / SH["or"])
        )
        if or_values:
            # or clauses share the variable name. Get the variablen name from the SH.name
            # or RDFS.label for the current pshape, or generate a new one
            or_var = get_varname(pshape)
            or_var = f"?{or_var}".replace(" ", "_")
            # connect ?target to the variable that will be used in the OR clauses
            clauses += f"{root_var} {path} {or_var} .\n"
            items = list(graph.objects(or_values, (RDF.rest * ZeroOrMore) / RDF.first))
            or_parts = []
            for item in items:
                or_body, or_project = _shape_to_where(graph, item, or_var)
                or_parts.append(or_body)
                project.update(or_project)
            clauses += " UNION ".join(f"{{ {or_body} }}" for or_body in or_parts)

        pvalue = graph.value(pshape, SH.hasValue)
        if pvalue:
            clauses += f"{root_var} {path} {pvalue.n3()} .\n"

        if not pclass and not pnode and not or_values and not pvalue:
            clauses += f"{root_var} {path} {name} .\n"

    return clauses, list(project)


def _resolve_imports(
    graph: rdflib.Graph,
    recursive_limit: int,
    seen: Set[rdflib.URIRef],
    error_on_missing_imports: bool = True,
) -> rdflib.Graph:
    from buildingmotif.dataclasses.library import Library

    bm = get_building_motif()

    logger = logging.getLogger(__name__)

    if recursive_limit == 0:
        return graph
    new_g = copy_graph(graph)
    for ontology in graph.objects(predicate=OWL.imports):
        if ontology in seen:
            continue
        seen.add(ontology)

        # go find the graph definition from our libraries
        try:
            lib = Library.load(name=ontology)
            sc_to_add = lib.get_shape_collection()
        except Exception as e:
            logger.warning(
                "Could not resolve import of %s from Libraries (%s). Trying shape collections",
                ontology,
                e,
            )
            sc_to_add = None

        # search through our shape collections for a graph with the provided name
        if sc_to_add is None:
            for shape_collection in bm.table_connection.get_all_db_shape_collections():
                sc = ShapeCollection.load(shape_collection.id)
                if sc.graph_name == ontology:
                    sc_to_add = sc
                    break
            logger.warning(
                "Could not resolve import of %s from Libraries. Trying shape collections",
                ontology,
            )

        if sc_to_add is None:
            if error_on_missing_imports:
                raise Exception("Could not resolve import of %s", ontology)
            continue

        dependency = _resolve_imports(
            sc_to_add.graph,
            recursive_limit - 1,
            seen,
            error_on_missing_imports=error_on_missing_imports,
        )
        new_g += dependency
    return new_g



from functools import cached_property
from typing import List

from buildingmotif.ingresses.base import Record, RecordIngressHandler
from buildingmotif.label_parsing import (
    Parser,
    analyze_failures,
    parse_list,
    results_to_tokens,
)


class NamingConventionIngress(RecordIngressHandler):
    """
    Ingress handler that parses labels using a naming convention parser.
    This returns a Record for each input label, with the parsed tokens as a field.
    You will need to attach this to a GraphIngressHandler to turn the tokens into a graph.
    """

    def __init__(
        self,
        upstream: RecordIngressHandler,
        naming_convention: Parser,
    ):
        """
        Create a new NamingConventionIngress.

        :param upstream: The upstream ingress handler to get records from. This should return records with a "label" field.
        :type upstream: RecordIngressHandler
        :param naming_convention: The naming convention parser to use.
        :type naming_convention: Parser
        """
        self.upstream = upstream
        self.naming_convention = naming_convention

    def dump_failed_labels(self):
        sorted_groups = sorted(
            analyze_failures(self.failures).items(),
            key=lambda x: len(x[1]),
            reverse=True,
        )
        for group, failures in sorted_groups:
            print(f"Unparsed label: {group} ({len(failures)} failures)")
            for failure in failures:
                print(f"\t{failure}")
            print()

    @cached_property
    def records(self) -> List[Record]:
        results, failures = parse_list(
            self.naming_convention, [x.fields["label"] for x in self.upstream.records]
        )
        self.failures = failures
        self.results = results
        tokens = results_to_tokens(results)
        return [
            Record(
                rtype="token",
                fields={
                    "label": t["label"],
                    "tokens": t["tokens"],
                },
            )
            for t in tokens
        ]



# configure logging output
import logging
import warnings
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple

try:
    import BAC0
    from BAC0.core.devices.Device import Device as BACnetDevice
except ImportError:
    logging.critical(
        "Install the 'bacnet-ingress' module, e.g. 'pip install buildingmotif[bacnet-ingress]'"
    )


from buildingmotif.ingresses.base import Record, RecordIngressHandler

# We do this little rigamarole to avoid BAC0 spitting out a million
# logging messages warning us that we changed the log level, which
# happens when we go through the normal BAC0 log level procedure
logger = logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base")
logger.setLevel(logging.ERROR)


class BACnetNetwork(RecordIngressHandler):
    def __init__(self, ip: Optional[str] = None):
        """
        Reads a BACnet network to discover the devices and objects therein

        :param ip: IP/mask for the host which is canning the networks,
                    defaults to None
        :type ip: Optional[str], optional
        """
        # create the network object; this will handle scans
        # Be a good net citizen: do not ping BACnet devices
        self.network = BAC0.connect(ip=ip, ping=False)
        # initiate discovery of BACnet networks
        self.network.discover()

        self.devices: List[BACnetDevice] = []
        self.objects: Dict[Tuple[str, int], List[dict]] = {}

        # for each discovered Device, create a BAC0.device object
        # This will read the BACnet objects off of the Device.
        # Save the BACnet objects in the objects dictionary
        try:
            if self.network.discoveredDevices is None:
                warnings.warn("BACnet ingress could not find any BACnet devices")
            for (address, device_id) in self.network.discoveredDevices:  # type: ignore
                # set poll to 0 to avoid reading the points regularly
                dev = BAC0.device(address, device_id, self.network, poll=0)
                self.devices.append(dev)
                self.objects[(address, device_id)] = []

                for bobj in dev.points:
                    obj = bobj.properties.asdict
                    self._clean_object(obj)
                    self.objects[(address, device_id)].append(obj)
        finally:
            for dev in self.devices:
                self.network.unregister_device(dev)
            self.network.disconnect()

    def _clean_object(self, obj: Dict[str, Any]):
        if "name" in obj:
            # remove trailing/leading whitespace from names
            obj["name"] = obj["name"].strip()

    @cached_property
    def records(self) -> List[Record]:
        """
        Returns a list of the BACnet devices and objects discovered in the
        BACnet network. The 'rtype' field of each Record is either "Device"
        for a BACnet Device or "Object" for a BACnet Object. The 'fields'
        field contains the key-value BACnet properties associated with that
        device or object.

        :return: list of BACnet devices and objects, each expressed as a Record
        :rtype: List[Record]
        """
        records = []
        # make devices
        for (address, device_id) in self.objects.keys():
            records.append(
                Record(
                    rtype="Device",
                    fields={"address": address, "device_id": device_id},
                )
            )
        for (address, device_id), objs in self.objects.items():
            for obj in objs:
                fields = obj.copy()
                del fields["device"]
                fields["device_id"] = device_id
                records.append(
                    Record(
                        rtype="Object",
                        fields=fields,
                    )
                )
        return records



from typing import List, Union

from typing_extensions import NotRequired, TypedDict

from buildingmotif.database.tables import DBTemplate
from buildingmotif.dataclasses.template import Template

TemplateDict = TypedDict(
    "TemplateDict",
    {
        "id": int,
        "name": str,
        "body_id": str,
        "optional_args": List[str],
        "library_id": int,
        "dependency_ids": List[int],
        "parameters": NotRequired[List[str]],
    },
)


def serialize(
    param: Union[DBTemplate, List[DBTemplate]], include_parameters: bool = False
) -> Union[TemplateDict, List[TemplateDict]]:
    """Serialize one or more templates into a TypedDict.

    :param param: one template or a list of templates.
    :type param: Union[DBTemplate, List[DBTemplate]]
    :param include_parameters: to include parameters, default False
    :type include_parameters: bool
    :raises ValueError: if invalid input
    :return: one JSON per serialized template
    :rtype: Union[TemplateDict, List[TemplateDict]]
    """
    if isinstance(param, DBTemplate):
        return _serialize(param, include_parameters)

    elif isinstance(param, list):
        return [_serialize(x, include_parameters) for x in param]

    raise ValueError("invalid input. Must be a DBTemplate or list of DBTemplates")


def _serialize(template: DBTemplate, include_parameters: bool = False) -> TemplateDict:
    """Serialize a template into a TypedDict.

    :param template: template
    :type template: DBTemplate
    :param include_parameters: to include parameters, default False
    :type include_parameters: bool
    :return: serialized template
    :rtype: TemplateDict
    """
    res: TemplateDict = {
        "id": template.id,
        "name": template.name,
        "body_id": template.body_id,
        "optional_args": template.optional_args,
        "library_id": template.library_id,
        "dependency_ids": [d.id for d in template.dependencies],
    }

    if include_parameters:
        parameters = Template.load(template.id).parameters
        res["parameters"] = sorted(
            list(parameters)
        )  # make it serzialiable and determintistic

    return res



import logging

from buildingmotif.ingresses.base import Record  # noqa
from buildingmotif.ingresses.csvingress import CSVIngress  # noqa
from buildingmotif.ingresses.naming_convention import NamingConventionIngress  # noqa
from buildingmotif.ingresses.template import (  # noqa
    TemplateIngress,
    TemplateIngressWithChooser,
)

try:
    from buildingmotif.ingresses.brick import (  # noqa
        BACnetNetwork,
        BACnetToBrickIngress,
    )
except ImportError as e:
    logging.warning(f"Could not import BACnet ingresses: {e}")



from typing import Dict, List

from sqlalchemy import Column, ForeignKey, Integer, String, Text, UniqueConstraint
from sqlalchemy.orm import Mapped, declarative_base, relationship

# from sqlalchemy.dialects.postgresql import JSON
from buildingmotif.database.utils import JSONType

Base = declarative_base()


class DBModel(Base):
    """A Model is a metadata model of all or part of a building."""

    __tablename__ = "models"
    id: Mapped[int] = Column(Integer, primary_key=True)
    name: Mapped[str] = Column(String())
    description: Mapped[str] = Column(Text(), default="", nullable=False)
    graph_id: Mapped[str] = Column(String())
    manifest_id: Mapped[int] = Column(
        Integer, ForeignKey("shape_collection.id"), nullable=False
    )
    manifest: "DBShapeCollection" = relationship(
        "DBShapeCollection",
        uselist=False,
        cascade="all,delete",
    )


class DBShapeCollection(Base):
    """A ShapeCollection is a collection of shapes, which are used to validate
    parts of a model.
    """

    __tablename__ = "shape_collection"
    id: Mapped[int] = Column(Integer, primary_key=True)
    graph_id: Mapped[str] = Column(String())


class DBLibrary(Base):
    """A Library is a distributable collection of Templates and Shapes."""

    __tablename__ = "library"
    id: Mapped[int] = Column(Integer, primary_key=True)
    name: Mapped[str] = Column(String(), nullable=False, unique=True)

    templates: Mapped[List["DBTemplate"]] = relationship(
        "DBTemplate", back_populates="library", cascade="all,delete"
    )

    shape_collection_id = Column(
        Integer, ForeignKey("shape_collection.id"), nullable=False
    )
    shape_collection: DBShapeCollection = relationship(
        "DBShapeCollection",
        uselist=False,
        cascade="all,delete",
    )


class DepsAssociation(Base):
    """Many-to-many relationship between dependant templates."""

    __tablename__ = "deps_association_table"

    id: Mapped[int] = Column(Integer, primary_key=True)
    dependant_id: Mapped[int] = Column(ForeignKey("template.id"))
    dependee_id: Mapped[int] = Column(ForeignKey("template.id"))
    # args are a mapping of dependee args to dependant args
    args: Mapped[Dict[str, str]] = Column(JSONType)  # type: ignore

    __table_args__ = (
        UniqueConstraint(
            "dependant_id",
            "dependee_id",
            "args",
            name="deps_association_unique_constraint",
        ),
    )


class DBTemplate(Base):
    """A Template is used to generate content for a model."""

    # TODO: doc table properties better.
    __tablename__ = "template"
    id: Mapped[int] = Column(Integer, primary_key=True)
    name: Mapped[str] = Column(String(), nullable=False)
    body_id: Mapped[str] = Column(String())
    optional_args: Mapped[List[str]] = Column(JSONType)  # type: ignore

    library_id: Mapped[int] = Column(Integer, ForeignKey("library.id"), nullable=False)
    library: Mapped[DBLibrary] = relationship("DBLibrary", back_populates="templates")
    dependencies: Mapped[List["DBTemplate"]] = relationship(
        "DBTemplate",
        secondary="deps_association_table",
        primaryjoin=id == DepsAssociation.dependant_id,
        secondaryjoin=id == DepsAssociation.dependee_id,
        back_populates="dependants",
    )
    dependants: Mapped[List["DBTemplate"]] = relationship(
        "DBTemplate",
        secondary="deps_association_table",
        primaryjoin=id == DepsAssociation.dependee_id,
        secondaryjoin=id == DepsAssociation.dependant_id,
        back_populates="dependencies",
    )

    __table_args__ = (
        UniqueConstraint(
            "name",
            "library_id",
            name="name_library_unique_constraint",
        ),
    )



import logging
from typing import Callable, Optional

from rdflib import BNode, Graph, Literal, Namespace, URIRef
from rdflib.term import Node

from buildingmotif.dataclasses import Template
from buildingmotif.ingresses.base import (
    GraphIngressHandler,
    Record,
    RecordIngressHandler,
)


class TemplateIngress(GraphIngressHandler):
    """
    Reads records and attempts to instantiate the given template
    with each record. Produces a graph.

    If 'inline' is True, inlines all templates when they are instantiated.
    """

    def __init__(
        self,
        template: Template,
        mapper: Optional[Callable[[str], str]],
        upstream: RecordIngressHandler,
        fill_unused: bool = False,
        inline: bool = False,
        require_optional_args: bool = True,
    ):
        """
        Create a new TemplateIngress handler

        :param template: the template to instantiate on each record
        :type template: Template
        :param mapper: Function which takes a column name as input and returns
                       the name of the parameter the corresponding cell should
                       be bound to. If None, uses the column name as the
                       parameter name
        :type mapper: Optional[Callable[[str], str]]
        :param upstream: the ingress handler from which to source records
        :type upstream: RecordIngressHandler
        :param inline: if True, inline the template before evaluating it on
                      each row, defaults to False
        :type inline: bool, optional
        :param require_optional_args: if True, require that optional arguments in the
                      chosen template be provided by the upstream ingress handler,
                      defaults to False
        :type require_optional_args: bool, optional
        :param fill_unused: if True, mint URIs for any unbound parameters in
                      the template for each input from the upstream ingress handler,
                      defaults to False
        :type fill_unused: bool, optional
        """
        self.mapper = mapper if mapper else lambda x: x
        self.upstream = upstream
        self.require_optional_args = require_optional_args
        if inline:
            self.template = template.inline_dependencies()
        else:
            self.template = template
        self.fill_unused = fill_unused

    def graph(self, ns: Namespace) -> Graph:
        g = Graph()

        # ensure 'ns' is a Namespace or URI forming won't work
        if not isinstance(ns, Namespace):
            ns = Namespace(ns)

        records = self.upstream.records
        assert records is not None
        for rec in records:
            bindings = {self.mapper(k): _get_term(v, ns) for k, v in rec.fields.items()}
            graph = self.template.evaluate(
                bindings, require_optional_args=self.require_optional_args
            )
            # if it is a graph then all expected params were provided and we are done!
            if isinstance(graph, Graph):
                g += graph
                continue
            # here, we know that the 'graph' variable is actually a Template. If fill_unused
            # is True, we use 'fill' on the template to generate a new graph
            if self.fill_unused:
                _, graph = graph.fill(ns, include_optional=self.require_optional_args)
                g += graph
                continue
            raise Exception(f"Paramaters {graph.parameters} are still unused!")
        return g


class TemplateIngressWithChooser(GraphIngressHandler):
    """
    Reads records and attempts to instantiate a template with each record.
    Uses a 'chooser' function to determine which template should be
    instantiated for each record. Produces a graph.

    If 'inline' is True, inlines all templates when they are instantiated.
    """

    def __init__(
        self,
        chooser: Callable[[Record], Template],
        mapper: Optional[Callable[[str], str]],
        upstream: RecordIngressHandler,
        inline=False,
        require_optional_args: bool = True,
    ):
        """
        Create a new TemplateIngress handler

        :param template: the template to instantiate on each record
        :type template: Template
        :param chooser: Function which takes a record and returns the Template
                       which the record should be evaluated on.
        :type mapper: Optional[Callable[[Record], Template]]
        :param mapper: Function which takes a column name as input and returns
                       the name of the parameter the corresponding cell should
                       be bound to. If None, uses the column name as the
                       parameter name
        :type mapper: Optional[Callable[[str], str]]
        :param upstream: the ingress handler from which to source records
        :type upstream: RecordIngressHandler

        :param inline: if True, inline the template before evaluating it on
                      each row, defaults to False
        :type inline: bool, optional
        :param require_optional_args: if True, require that optional arguments in the
                      chosen template be provided by the upstream ingress handler,
                      defaults to False
        :type require_optional_args: bool, optional
        """
        self.chooser = chooser
        self.mapper = mapper if mapper else lambda x: x
        self.upstream = upstream
        self.inline = inline
        self.require_optional_args = require_optional_args

    def graph(self, ns: Namespace) -> Graph:
        g = Graph()

        # ensure 'ns' is a Namespace or URI forming won't work
        if not isinstance(ns, Namespace):
            ns = Namespace(ns)

        records = self.upstream.records
        assert records is not None
        for rec in records:
            template = self.chooser(rec)
            if template is None:
                logging.warning(f"Chooser function does not give a template for {rec}")
                continue
            if self.inline:
                template = template.inline_dependencies()
            bindings = {self.mapper(k): _get_term(v, ns) for k, v in rec.fields.items()}
            graph = template.evaluate(
                bindings, require_optional_args=self.require_optional_args
            )
            if not isinstance(graph, Graph):
                _, graph = graph.fill(ns)
            g += graph
        return g


def _get_term(field_value: str, ns: Namespace) -> Node:
    assert isinstance(ns, Namespace), f"{ns} must be a rdflib.Namespace instance"
    if isinstance(field_value, (URIRef, Literal, BNode)):
        return field_value
    try:
        uri = URIRef(ns[field_value])
        uri.n3()  # raises an exception if invalid URI
        return uri
    except Exception:
        return Literal(field_value)



from buildingmotif.label_parsing.parser import (
    Parser,
    analyze_failures,
    parse_list,
    results_to_tokens,
)
from buildingmotif.label_parsing.tokens import (
    Constant,
    Delimiter,
    ErrorTokenResult,
    Identifier,
    Null,
    Token,
    TokenOrConstructor,
    TokenResult,
)

__all__ = [
    "Parser",
    "analyze_failures",
    "parse_list",
    "results_to_tokens",
    "Token",
    "TokenResult",
    "TokenOrConstructor",
    "Identifier",
    "Constant",
    "Delimiter",
    "Null",
    "ErrorTokenResult",
]



from rdflib import Namespace

BMOTIF = Namespace("https://nrel.gov/BuildingMOTIF#")

# special namespace to denote template parameters inside RDF graphs
PARAM = Namespace("urn:___param___#")

# all versions of Brick > 1.1 have these namespaces
BRICK = Namespace("https://brickschema.org/schema/Brick#")
TAG = Namespace("https://brickschema.org/schema/BrickTag#")
BSH = Namespace("https://brickschema.org/schema/BrickShape#")
REF = Namespace("https://brickschema.org/schema/Brick/ref#")

# defaults
OWL = Namespace("http://www.w3.org/2002/07/owl#")
RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
SH = Namespace("http://www.w3.org/ns/shacl#")
XSD = Namespace("http://www.w3.org/2001/XMLSchema#")

# QUDT namespaces
QUDT = Namespace("http://qudt.org/schema/qudt/")
QUDTQK = Namespace("http://qudt.org/vocab/quantitykind/")
QUDTDV = Namespace("http://qudt.org/vocab/dimensionvector/")
UNIT = Namespace("http://qudt.org/vocab/unit/")

# ASHRAE namespaces
BACNET = Namespace("http://data.ashrae.org/bacnet/2020#")
S223 = Namespace("http://data.ashrae.org/standard223#")

BM = Namespace("https://nrel.gov/BuildingMOTIF#")
CONSTRAINT = Namespace("https://nrel.gov/BuildingMOTIF/constraints#")
S223 = Namespace("http://data.ashrae.org/standard223#")

A = RDF.type


def bind_prefixes(graph):
    """Associate common prefixes with the graph.

    :param graph: graph
    :type graph: rdflib.Graph
    """
    graph.bind("xsd", XSD)
    graph.bind("rdf", RDF)
    graph.bind("owl", OWL)
    graph.bind("rdfs", RDFS)
    graph.bind("skos", SKOS)
    graph.bind("sh", SH)
    graph.bind("qudtqk", QUDTQK)
    graph.bind("qudt", QUDT)
    graph.bind("unit", UNIT)
    graph.bind("brick", BRICK)
    graph.bind("tag", TAG)
    graph.bind("bsh", BSH)
    graph.bind("P", PARAM)
    graph.bind("constraint", CONSTRAINT)
    graph.bind("bmotif", BM)



import logging
import secrets
from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

import pyshacl  # type: ignore
from rdflib import BNode, Graph, Literal, URIRef
from rdflib.compare import _TripleCanonicalizer
from rdflib.paths import ZeroOrOne
from rdflib.term import Node

from buildingmotif.database.errors import TemplateNotFound
from buildingmotif.namespaces import OWL, PARAM, RDF, SH, XSD, bind_prefixes

if TYPE_CHECKING:
    from buildingmotif.dataclasses import Library, Template

Triple = Tuple[Node, Node, Node]
_gensym_counter = 0


def _strip_param(param: Union[Node, str]) -> str:
    """
    Strips all PARAM namespaces from the input parameter
    """
    param = str(param)
    while param.startswith(PARAM):
        param = param[len(PARAM) :]
    return param


def _gensym(prefix: str = "p") -> URIRef:
    """
    Generate a unique identifier.
    """
    global _gensym_counter
    _gensym_counter += 1
    return PARAM[f"{prefix}{_gensym_counter}"]


def _param_name(param: URIRef) -> str:
    """
    Returns the name of the param without the prefix.
    """
    assert param.startswith(PARAM)
    return param[len(PARAM) :]


def _guarantee_unique_template_name(library: "Library", name: str) -> str:
    """
    Ensure that the template name is unique in the library by appending an increasing
    number. This is only called when we are generating templates from GraphDiffs and
    the names are local to the Library. The Library is intended to be ephemeral so these
    names will not be around for long.
    """
    idx = 1
    original_name = name
    try:
        while library.get_template_by_name(name):
            name = f"{original_name}_{idx}"
            idx += 1
    except TemplateNotFound:
        # this means that the template does not exist and we can use the original name
        pass
    return name


def copy_graph(g: Graph, preserve_blank_nodes: bool = True) -> Graph:
    """
    Copy a graph. Creates new blank nodes so that these remain unique to each Graph

    :param g: the graph to copy
    :type g: Graph
    :param preserve_blank_nodes: if true, keep blank nodes the same when copying the graph
    :type preserve_blank_nodes: boolean, defaults to True
    :return: a copy of the input graph
    :rtype: Graph
    """
    c = Graph()
    new_prefix = secrets.token_hex(4)
    for t in g.triples((None, None, None)):
        assert isinstance(t, tuple)
        (s, p, o) = t
        if not preserve_blank_nodes:
            if isinstance(s, BNode):
                s = BNode(value=new_prefix + s.toPython())
            if isinstance(p, BNode):
                p = BNode(value=new_prefix + p.toPython())
            if isinstance(o, BNode):
                o = BNode(value=new_prefix + o.toPython())
        c.add((s, p, o))
    return c


def inline_sh_nodes(g: Graph):
    """
    Recursively inlines all sh:node properties and objects on the graph.

    :param g: graph to be edited
    :type g: Graph
    """
    q = """
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    CONSTRUCT {
        ?parent ?p ?o .
    }
    WHERE {
        ?parent sh:node ?child .
        ?child ?p ?o
    }"""
    original_size = 0
    while original_size != len(g):  # type: ignore
        original_size = len(g)  # type: ignore
        for (s, p, o) in g.query(q):  # type: ignore
            if p == RDF.type and o == SH.NodeShape:
                continue
            g.add((s, p, o))
        break


def combine_graphs(*graphs: Graph) -> Graph:
    """Combine all of the graphs into a new graph.

    :return: combined graph
    :rtype: Graph
    """
    newg = Graph()
    for graph in graphs:
        newg += graph
    return newg


def graph_size(g: Graph) -> int:
    """Returns the number of triples in a graph.

    :param g: graph to be measured
    :type g: Graph
    :return: number of triples in the graph
    :rtype: int
    """
    return len(tuple(g.triples((None, None, None))))


def remove_triples_with_node(g: Graph, node: URIRef) -> None:
    """Remove all triples that include the given node. Edits the graph
    in-place.

    :param g: the graph to remove triples from
    :type g: Graph
    :param node: the node to remove
    :type node: URIRef
    """
    for triple in g.triples((node, None, None)):
        g.remove(triple)
    for triple in g.triples((None, node, None)):
        g.remove(triple)
    for triple in g.triples((None, None, node)):
        g.remove(triple)


def replace_nodes(g: Graph, replace: Dict[Node, Node]) -> None:
    """Replace nodes in a graph.

    :param g: graph to replace nodes in
    :type g: Graph
    :param replace: dict mapping old nodes to new nodes
    :type replace: Dict[Node, Node]
    """
    for s, p, o in g.triples((None, None, None)):
        g.remove((s, p, o))
        if s in replace:
            s = replace[s]
        if p in replace:
            p = replace[p]
        if o in replace:
            o = replace[o]
        g.add((s, p, o))


def get_ontology_files(directory: Path, recursive: bool = True) -> List[Path]:
    """Returns a list of all ontology files in the given directory.

    If recursive is true, traverses the directory structure to find ontology
    files not just in the given directory.

    :param directory: the directory to begin the search
    :type directory: Path
    :param recursive: if true, find ontology files in nested directories,
        defaults to true
    :type recursive: bool, optional
    :return: a list of filenames
    :rtype: List[Path]
    """
    patterns = [
        "*.ttl",
        "*.n3",
        "*.rdf",
        "*.xml",
    ]
    if recursive:
        searches = (directory.rglob(f"{pat}") for pat in patterns)
    else:
        searches = (directory.glob(f"{pat}") for pat in patterns)
    # filter out files in .ipynb_checkpoints
    filtered_searches = (
        filter(lambda x: ".ipynb_checkpoints" not in Path(x).parts, search)
        for search in searches
    )
    return list(chain.from_iterable(filtered_searches))


def get_template_parts_from_shape(
    shape_name: URIRef, shape_graph: Graph
) -> Tuple[Graph, List[Dict]]:
    """Turn a SHACL shape into a template. The following attributes of
    NodeShapes will be incorporated into the resulting template:
    - sh:property with sh:minCount
    - sh:property with sh:qualifiedMinCount
    - sh:class
    - sh:node

    :param shape_name: name of shape
    :type shape_name: URIRef
    :param shape_graph: shape graph
    :type shape_graph: Graph
    :raises Exception: if more than one object type detected on shape
    :raises Exception: if more than one min count detected on shape
    :return: template parts
    :rtype: Tuple[Graph, List[Dict]]
    """
    # TODO: sh:or?
    # the template body
    body = Graph()
    root_param = PARAM["name"]

    deps = []

    pshapes = shape_graph.objects(subject=shape_name, predicate=SH["property"])
    for pshape in pshapes:
        property_path = shape_graph.value(pshape, SH["path"])
        if property_path is None:
            raise Exception(
                f"no sh:path detected on {shape_name} property shape {pshape}"
            )
        # TODO: expand otypes to include sh:in, sh:or, or no datatype at all!
        otypes = list(
            shape_graph.objects(
                subject=pshape,
                predicate=SH["qualifiedValueShape"]
                * ZeroOrOne  # type:ignore
                / (SH["class"] | SH["node"] | SH["datatype"]),
            )
        )
        mincounts = list(
            shape_graph.objects(
                subject=pshape, predicate=SH["minCount"] | SH["qualifiedMinCount"]
            )
        )
        if len(otypes) > 1:
            raise Exception(f"more than one object type detected on {shape_name}")
        if len(mincounts) > 1:
            raise Exception(f"more than one min count detected on {shape_name}")
        if len(mincounts) == 0 or len(otypes) == 0:
            # print(f"No useful information on {shape_name} - {pshape}")
            # print(shape_graph.cbd(pshape).serialize())
            continue
        (path, otype, mincount) = property_path, otypes[0], mincounts[0]
        assert isinstance(mincount, Literal)

        param_name = shape_graph.value(pshape, SH["name"])

        for num in range(int(mincount)):
            if param_name is not None:
                param = PARAM[f"{param_name}{num}"]
            else:
                param = _gensym()
            body.add((root_param, path, param))

            otype_as_class = (None, SH["class"], otype) in shape_graph
            otype_as_node = (None, SH["node"], otype) in shape_graph
            otype_is_nodeshape = (otype, RDF.type, SH.NodeShape) in shape_graph

            if (otype_as_class and otype_is_nodeshape) or otype_as_node:
                if not isinstance(otype, URIRef):
                    continue
                deps.append({"template": str(otype), "args": {"name": param}})
                body.add((param, RDF.type, otype))

        pvalue = shape_graph.value(pshape, SH["hasValue"])
        if pvalue:
            body.add((root_param, path, pvalue))

    if (shape_name, RDF.type, OWL.Class) in shape_graph:
        body.add((root_param, RDF.type, shape_name))

    classes = shape_graph.objects(shape_name, SH["class"])
    for cls in classes:
        body.add((root_param, RDF.type, cls))

    classes = shape_graph.objects(shape_name, SH["targetClass"])
    for cls in classes:
        body.add((root_param, RDF.type, cls))

    # for all objects of sh:node, add them to the deps if they haven't been added
    # already through the property shapes above
    nodes = shape_graph.cbd(shape_name).objects(predicate=SH["node"], unique=True)
    for node in nodes:
        # if node is already in deps, skip it
        if any(str(node) == dep["template"] for dep in deps):
            continue
        # skip non-URIRef nodes
        if not isinstance(node, URIRef):
            continue
        deps.append(
            {"template": str(node), "args": {"name": "name"}}
        )  # tie to root param

    return body, deps


@dataclass
class _TemplateIndex:
    template: "Template"
    param_types: Dict[Node, List[Node]]
    prop_types: Dict[Node, List[Node]]
    prop_values: Dict[Node, List[Node]]
    prop_shapes: Dict[Node, List[Node]]
    target: URIRef

    @property
    def target_type(self):
        return self.param_types[self.target][0]


def _prep_shape_graph() -> Graph:
    shape = Graph()
    bind_prefixes(shape)
    shape.bind("mark", PARAM)
    return shape


def _index_properties(templ: "Template") -> _TemplateIndex:
    templ_graph = templ.evaluate(
        {p: PARAM[p] for p in templ.parameters}, {"mark": PARAM}
    )
    assert isinstance(templ_graph, Graph)

    # pick a random node to act as the 'target' of the shape
    target = next(iter(templ_graph.subjects(RDF.type)))
    print(f"Choosing {target} as the target of the shape")
    assert isinstance(target, URIRef)

    # store the classes for each parameter
    param_types: Dict[Node, List[Node]] = defaultdict(list)
    for (param, ptype) in templ_graph.subject_objects(RDF.type):
        param_types[param].append(ptype)

    # store the properties and their types for the target
    prop_types: Dict[Node, List[Node]] = defaultdict(list)
    prop_values: Dict[Node, List[Node]] = defaultdict(list)
    prop_shapes: Dict[Node, List[Node]] = defaultdict(list)
    # TODO: prop_shapes for all properties whose object corresponds to another shape
    for p, o in templ_graph.predicate_objects(target):
        if p == RDF.type:
            continue
        # maybe_param = str(o).removeprefix(PARAM) Python >=3.9
        maybe_param = str(o)[len(PARAM) :]
        if maybe_param in templ.dependency_parameters:
            dep = templ.dependency_for_parameter(maybe_param)
            if dep is not None:
                prop_shapes[p].append(URIRef(dep._name))
        elif o in param_types:
            prop_types[p].append(param_types[o][0])
        elif str(o) not in PARAM:
            prop_values[p].append(o)
        elif str(o) in PARAM and o not in param_types:
            logging.warn(
                f"{o} is does not have a type and does not seem to be a literal"
            )
    return _TemplateIndex(
        templ,
        dict(param_types),
        dict(prop_types),
        dict(prop_values),
        dict(prop_shapes),
        target,
    )


def _add_property_shape(
    graph: Graph, name: Node, constraint: Node, path: Node, value: Node
):
    pshape = BNode()
    graph.add((name, SH.property, pshape))
    graph.add((pshape, SH.path, path))
    graph.add((pshape, constraint, value))
    graph.add((pshape, SH["minCount"], Literal(1)))
    graph.add((pshape, SH["maxCount"], Literal(1)))


def _add_qualified_property_shape(
    graph: Graph, name: Node, constraint: Node, path: Node, value: Node
):
    pshape = BNode()
    graph.add((name, SH.property, pshape))
    graph.add((pshape, SH.path, path))
    qvc = BNode()
    graph.add((pshape, SH["qualifiedValueShape"], qvc))
    graph.add((qvc, constraint, value))
    graph.add((pshape, SH["qualifiedMinCount"], Literal(1)))
    graph.add((pshape, SH["qualifiedMaxCount"], Literal(1)))


def template_to_shape(template: "Template") -> Graph:
    """Turn this template into a SHACL shape.

    :param template: template to convert
    :type template: template
    :return: graph of template
    :rtype: Graph
    """
    # TODO If 'use_all' is True, this will create a shape that incorporates all
    # Templates by the same name in the same Library.
    templ = copy(template)
    shape = _prep_shape_graph()

    idx = _index_properties(templ)

    shape.add((PARAM[templ.name], SH.targetClass, idx.target_type))
    # create the shape
    shape.add((PARAM[templ.name], RDF.type, SH.NodeShape))
    shape.add((PARAM[templ.name], SH["class"], idx.target_type))
    for prop, ptypes in idx.prop_types.items():
        if len(ptypes) == 1:
            _add_property_shape(shape, PARAM[templ.name], SH["class"], prop, ptypes[0])
        else:  # more than one ptype
            for ptype in ptypes:
                _add_qualified_property_shape(
                    shape, PARAM[templ.name], SH["class"], prop, ptype
                )
    for prop, values in idx.prop_values.items():
        if len(values) == 1:
            _add_property_shape(shape, PARAM[templ.name], SH.hasValue, prop, values[0])
        else:  # more than one ptype
            for value in values:
                _add_qualified_property_shape(
                    shape, PARAM[templ.name], SH.hasValue, prop, value
                )
    for prop, shapes in idx.prop_shapes.items():
        if len(shapes) == 1:
            _add_property_shape(shape, PARAM[templ.name], SH["node"], prop, shapes[0])
        else:  # more than one ptype
            for shp in shapes:
                _add_qualified_property_shape(
                    # TODO: fix this?
                    shape,
                    PARAM[templ.name],
                    SH.node,
                    prop,
                    PARAM[str(shp)],
                )

    return shape


def new_temporary_graph(more_namespaces: Optional[dict] = None) -> Graph:
    """Creates a new in-memory RDF graph with common and additional namespace
    bindings.

    :param more_namespaces: namespaces, defaults to None
    :type more_namespaces: Optional[dict], optional
    :return: graph with namespaces
    :rtype: Graph
    """
    g = Graph()
    bind_prefixes(g)
    if more_namespaces:
        for prefix, uri in more_namespaces.items():
            g.bind(prefix, uri)
    return g


def get_parameters(graph: Graph) -> Set[str]:
    """Returns the set of parameter names in the given graph.

    Parameters are identified by the PARAM namespace, `urn:___param___#`. This
    method returns the names of the parameters, not the full URIs. For example,
    the parameter `urn:___param___#abc` in a graph would be returned as `abc`.

    :param graph: a graph containing parameters
    :type graph: Graph
    :return: a set of the parameter names in the graph
    :rtype: Set[str]
    """
    # get an iterable of all nodes in the graph
    all_nodes = chain.from_iterable(graph.triples((None, None, None)))
    # get all nodes in the PARAM namespace
    params = {str(node) for node in all_nodes if str(node).startswith(PARAM)}
    # extract the 'value' part of the param, which is the name of the parameter
    return {node[len(PARAM) :] for node in params}


def _inline_sh_node(sg: Graph):
    """
    This detects the use of 'sh:node' on SHACL NodeShapes and inlines
    the shape they point to.
    """
    q = """
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    SELECT ?parent ?child WHERE {
        ?parent a sh:NodeShape ;
                sh:node ?child .
        }"""
    for row in sg.query(q):
        parent, child = row  # type: ignore
        sg.remove((parent, SH.node, child))
        pos = sg.predicate_objects(child)
        for (p, o) in pos:
            sg.add((parent, p, o))


def _inline_sh_and(sg: Graph):
    """
    This detects the use of 'sh:and' on SHACL NodeShapes and inlines
    all of the included shapes
    """
    q = """
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    SELECT ?parent ?child ?andnode WHERE {
        ?parent a sh:NodeShape ;
                sh:and ?andnode .
        ?andnode rdf:rest*/rdf:first ?child .
        }"""
    for row in sg.query(q):
        parent, child, to_remove = row  # type: ignore
        sg.remove((parent, SH["and"], to_remove))
        pos = sg.predicate_objects(child)
        for (p, o) in pos:
            sg.add((parent, p, o))


def _inline_sh_qualified_value_shape(sg: Graph):
    """
    This detects the use of 'sh:qualifiedValueShape' on SHACL PropertyShapes and inlines
    all of the included shapes
    """
    q = """
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    SELECT ?parent ?child WHERE {
        ?parent a sh:PropertyShape ;
                sh:qualifiedValueShape ?child .
        }"""
    for row in sg.query(q):
        parent, child = row  # type: ignore
        sg.remove((parent, SH["qualifiedValueShape"], child))
        pobjs = sg.predicate_objects(child)
        for (p, o) in pobjs:
            sg.add((parent, p, o))


def rewrite_shape_graph(g: Graph) -> Graph:
    """
    Rewrites the input graph to make the resulting validation report more useful.

    :param g: the shape graph to rewrite
    :type g: Graph
    :return: a *copy* of the original shape graph w/ rewritten shapes
    :rtype: Graph
    """
    sg = copy_graph(g)

    previous_size = -1
    while len(sg) != previous_size:  # type: ignore
        previous_size = len(sg)  # type: ignore
        _inline_sh_and(sg)
        # make sure to handle sh:node *after* sh:and
        _inline_sh_node(sg)
        _inline_sh_qualified_value_shape(sg)
    return sg


def skip_uri(uri: URIRef) -> bool:
    """
    Returns True if the URI should be skipped during processing
    because it is axiomatic or not expected to be imported.

    Skips URIs in the XSD and SHACL namespaces.

    :return: True if the URI should be skipped; False otherwise
    :rtype: bool
    """
    # Now that we have all the templates, we can populate the dependencies.
    # IGNORES missing XSD imports --- there is really no reason to import the XSD
    # ontology because the handling is baked into the software processing the RDF
    # graph. Thus, XSD URIs will always yield import warnings. This is noisy, so we
    # suppress them.
    skip_ns = [XSD, SH]
    for ns in skip_ns:
        if uri.startswith(ns):
            return True
    return False


def shacl_validate(
    data_graph: Graph,
    shape_graph: Optional[Graph] = None,
    engine: Optional[str] = "topquadrant",
) -> Tuple[bool, Graph, str]:
    """
    Validate the data graph against the shape graph.
    Uses the fastest validation method available. Use the 'topquadrant' feature
    to use TopQuadrant's SHACL engine. Defaults to using PySHACL.

    :param data_graph: the graph to validate
    :type data_graph: Graph
    :param shape_graph: the shape graph to validate against
    :type shape_graph: Graph, optional
    :param engine: the SHACL engine to use, defaults to "topquadrant"
    :type engine: str, optional
    :return: a tuple containing the validation result, the validation report, and the validation report string
    :rtype: Tuple[bool, Graph, str]
    """

    if engine == "topquadrant":
        try:
            from brick_tq_shacl.topquadrant_shacl import (
                validate as tq_validate,  # type: ignore
            )

            return tq_validate(data_graph, shape_graph or Graph())  # type: ignore
        except ImportError:
            logging.info(
                "TopQuadrant SHACL engine not available. Using PySHACL instead."
            )
            pass

    data_graph = data_graph + (shape_graph or Graph())
    return pyshacl.validate(
        data_graph,
        shacl_graph=shape_graph,
        ont_graph=shape_graph,
        advanced=True,
        js=True,
        allow_warnings=True,
    )  # type: ignore


def shacl_inference(
    data_graph: Graph,
    shape_graph: Optional[Graph] = None,
    engine: Optional[str] = "topquadrant",
) -> Graph:
    """
    Infer new triples in the data graph using the shape graph.
    Edits the data graph in place. Uses the fastest inference method available.
    Use the 'topquadrant' feature to use TopQuadrant's SHACL engine. Defaults to
    using PySHACL.

    :param data_graph: the graph to infer new triples in
    :type data_graph: Graph
    :param shape_graph: the shape graph to use for inference
    :type shape_graph: Optional[Graph]
    :param engine: the SHACL engine to use, defaults to "topquadrant"
    :type engine: str, optional
    :return: the data graph with inferred triples
    :rtype: Graph
    """
    if engine == "topquadrant":
        try:
            from brick_tq_shacl.topquadrant_shacl import infer as tq_infer

            return tq_infer(data_graph, shape_graph or Graph())  # type: ignore
        except ImportError:
            logging.info(
                "TopQuadrant SHACL engine not available. Using PySHACL instead."
            )
            pass

    # We use a fixed-point computation approach to 'compiling' RDF models.
    # We accomlish this by keeping track of the size of the graph before and after
    # the inference step. If the size of the graph changes, then we know that the
    # inference has had some effect. We do this at most 3 times to avoid looping
    # forever.
    pre_compile_length = len(data_graph)  # type: ignore
    pyshacl.validate(
        data_graph=data_graph,
        shacl_graph=shape_graph,
        ont_graph=shape_graph,
        advanced=True,
        inplace=True,
        js=True,
        allow_warnings=True,
    )
    post_compile_length = len(data_graph)  # type: ignore

    attempts = 3
    while attempts > 0 and post_compile_length != pre_compile_length:
        pre_compile_length = len(data_graph)  # type: ignore
        pyshacl.validate(
            data_graph=data_graph,
            shacl_graph=shape_graph,
            ont_graph=shape_graph,
            advanced=True,
            inplace=True,
            js=True,
            allow_warnings=True,
        )
        post_compile_length = len(data_graph)  # type: ignore
        attempts -= 1
    return data_graph - (shape_graph or Graph())


def skolemize_shapes(g: Graph) -> Graph:
    """
    Skolemize the shapes in the graph.

    :param g: the graph to skolemize
    :type g: Graph
    :return: the skolemized graph
    :rtype: Graph
    """
    # write a query to update agraph by changing all PropertyShape blank nodes
    # to URIRefs
    g = copy_graph(g)
    property_shapes = list(g.subjects(predicate=RDF.type, object=SH.PropertyShape))
    property_shapes.extend(list(g.objects(predicate=SH.property)))
    replacements = {}
    for ps in property_shapes:
        # if not bnode, skip
        if not isinstance(ps, BNode):
            continue
        # create a new URIRef
        new_ps = URIRef(f"urn:well-known/{secrets.token_hex(4)}")
        # replace the old BNode with the new URIRef
        replacements[ps] = new_ps
    # apply the replacements
    replace_nodes(g, replacements)

    # name all objects of qualifiedValueShape
    qvs = list(g.objects(predicate=SH.qualifiedValueShape))
    replacements = {}
    for qv in qvs:
        # if not bnode, skip
        if not isinstance(qv, BNode):
            continue
        # create a new URIRef
        new_qv = URIRef(f"urn:well-known/{secrets.token_hex(4)}")
        # replace the old BNode with the new URIRef
        replacements[qv] = new_qv
    # apply the replacements
    replace_nodes(g, replacements)
    return g


def graph_hash(graph: Graph) -> int:
    """
    Returns a cryptographic hash of the graph contents.
    This uses the same method as rdflib's isomorphic function to generate a cryptographic hash of a given graph.
    This method calculates a consistent hash of the canonicalized form of the graph.
    If the hashes of two graphs are equal, this means that the graphs are isomorphic.
    Generating the hashes (using this method) and caching them allows graph isomorphism to be determined
    without having to recalculate the canonical form of the graph, which can be expensive.

    :param graph: graph to hash
    :type graph: graph

    :return: integer hash
    :rtype: int
    """
    # Copy graph to memory (improved performance if graph is backed by a DB store)
    graph_prime = copy_graph(graph)
    triple_canonicalizer = _TripleCanonicalizer(graph_prime)

    return triple_canonicalizer.to_hash()



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from inspect import Parameter, signature
from typing import Dict, List, Tuple

from buildingmotif.label_parsing.tokens import Constant, Identifier, Null, TokenResult

# TODO: programming by example?
# TODO: get LLM to write a parser for the point labels, given a human description


# type definition for the output of a parser function
@dataclass(frozen=True)
class ParseResult:
    tokens: List[TokenResult]
    success: bool
    _errors: List[str] = field(default_factory=list)

    @property
    def errors(self):
        """Return a list of errors and the offset into the string
        where the error occurred."""
        errors = []
        offset = 0
        for t in self.tokens:
            if t.error:
                errors.append((t.error, offset))
            offset += t.length
        return errors


# type definition for the parser functions.
# A parser function takes a string and returns a list of tuples
# each tuple is a token, the type of the token, and the length of the token
# the length of the token is used to keep track of how much of the string
# has been parsed
class Parser(ABC):
    __args__: dict

    def __new__(mcls, *args, **kwargs):
        """When a parser is constructed, save its arguments into a dictionary __args__.
        This allows parsers to be serialized later without requiring bespoke (de)serialization code
        for every parser type."""
        cls = super().__new__(mcls)
        sig = signature(mcls.__init__)
        parameters = sig.parameters
        arguments = sig.bind(cls, *args, **kwargs).arguments

        cls.__args__ = {}
        for name, value in arguments.items():
            if name != "self":
                kind = parameters[name].kind
                if kind in [
                    Parameter.POSITIONAL_ONLY,
                    Parameter.POSITIONAL_OR_KEYWORD,
                    Parameter.KEYWORD_ONLY,
                ]:
                    cls.__args__[name] = value
                elif kind == Parameter.VAR_POSITIONAL:
                    cls.__args__[name] = list(value)
                elif kind == Parameter.VAR_KEYWORD:
                    cls.__args__.update(value)

        return cls

    @abstractmethod
    def __call__(self, target: str) -> List[TokenResult]:
        pass


# function which takes the results of parse_list and turns all of the
# results (not failures) into token dictionaries.
# A token dictionary has an 'identifier' key which is the label (the keys
# of the results dictionary), and then a list of tokens. A token is a dictionary
# with 'identifier' (the substring part of the result) and 'type' (the type of
# the result if it is a constant)
def results_to_tokens(results):
    tokens = []
    for r in results:
        res = {"label": r, "tokens": []}
        parts = iter(results[r])
        first = None
        while True:
            try:
                # get first constant or identifier token using itertools
                first = first_true(
                    parts, pred=lambda x: isinstance(x.token, (Constant, Identifier))
                )
                # get the next constant or identifier token
                second = first_true(
                    parts, pred=lambda x: isinstance(x.token, (Constant, Identifier))
                )
                if not first or not second:
                    break
                # add the constant and identifier to the token dictionary
                identifier = first if isinstance(first.token, Identifier) else second
                constant = first if isinstance(first.token, Constant) else second
                res["tokens"].append(
                    {
                        "identifier": identifier.token.value,
                        "type": constant.token.value.toPython(),
                    }
                )
            except StopIteration:
                break
        if first is None:
            # if there are any constants left, add them to the token dictionary with the label
            first = first_true(parts, pred=lambda x: isinstance(x.token, Constant))
        if first:
            res["tokens"].append(
                {"identifier": r, "type": first.token.value.toPython()}
            )
        tokens.append(res)

    return tokens


# Analyzes the failures of a parser to capture all point labels.
# For each label in the failures, compute the length of the found tokens.
# Create a dictionary keyed with the label. The value should have two keys.
# The first key is the remaining "unparsed" part of the string, the second key
# is the set of tokens that were found.
def analyze_failures(failures: Dict[str, List[TokenResult]]):
    """Analyze the failures of a parser."""
    analyzed = {}
    for failure in failures:
        tokens = failures[failure]
        length = sum([t.length for t in tokens])
        analyzed[failure] = {
            "unparsed": failure[length:],
            "tokens": [
                {"identifier": t.token.value, "type": t.token.value} for t in tokens
            ],
        }
    # group the points by the unparsed portion
    grouped = defaultdict(list)
    for f in analyzed:
        grouped[analyzed[f]["unparsed"]].append(f)
    # for each group, add the tokens to the first point in the group
    return grouped


# wrapper function for reading a list of strings
# applies a given parser to each string in the list
# returns a dictionary of the input strings to the result of the parser
# Keep track of all strings that fail to parse and return them in a list
def parse_list(
    parser, target_list
) -> Tuple[Dict[str, List[TokenResult]], Dict[str, List[TokenResult]]]:
    """
    Parse a list of strings using the given parser.

    :param parser: the parsing combinator function
    :type parser: Parser
    :param target_list: the list of strings to parse
    :type target_list: List[str]
    :return: a tuple of the results and failures
    """
    results = {}
    failed = {}
    for target in target_list:
        result = parse(parser, target)
        if result.success:
            results[target] = result.tokens
        else:
            failed[target] = result.tokens
    return results, failed


# wrapper function for a parser that does the following:
# - apply the parser to the target
# - if the parser does not consume all the target, raise an error
# - return the result of the parser
def parse(parser: Parser, target: str) -> ParseResult:
    """
    Parse the given target string using the given parser.

    :param parser: the parsing combinator function
    :type parser: Parser
    :param target: the target string to parse
    :type target: str
    :return: the result of the parser
    :rtype: ParseResult
    """
    result = parser(target)
    # remove empty Null tokens from result
    result = [r for r in result if r.error or (not isinstance(r.token, Null))]
    # check length of target vs length of all results
    total_length = sum([r.length for r in result])
    return ParseResult(
        result, total_length == len(target), [r.error for r in result if r.error]
    )


# from itertools documentation
def first_true(iterable, default=None, pred=None):
    """Returns the first true value in the iterable.

    If no true value is found, returns *default*

    If *pred* is not None, returns the first item
    for which pred(item) is true.

    """
    # first_true([a,b,c], x) --> a or b or c or x
    # first_true([a,b], x, f) --> a if f(a) else b if f(b) else x
    return next(filter(pred, iterable), default)



from csv import DictReader
from functools import cached_property
from io import StringIO
from pathlib import Path
from typing import List, Optional, Union

from buildingmotif.ingresses.base import Record, RecordIngressHandler


class CSVIngress(RecordIngressHandler):
    """Reads rows from a CSV file and exposes them as records.
    The type of the record is the name of the CSV file
    """

    def __init__(
        self,
        filename: Optional[Path] = None,
        data: Optional[Union[str, StringIO]] = None,
    ):
        if filename is not None and data is not None:
            raise ValueError("Both filename and data are defined.")

        if filename:
            self.dict_reader = DictReader(open(filename))
            self.rtype = str(filename)

        elif data:
            # if data is a string, convert to StringIO
            if isinstance(data, str):
                data = StringIO(data)
            self.dict_reader = DictReader(data, delimiter=",")
            self.rtype = "data stream"

        else:
            raise ValueError("Either filename or data must be defined.")

    @cached_property
    def records(self) -> List[Record]:
        records = []
        for row in self.dict_reader:
            rec = Record(
                rtype=self.rtype,
                fields=row,
            )
            records.append(rec)

        return records



from io import StringIO
from typing import Dict

import flask
from flask import Blueprint, current_app, jsonify, request
from flask_api import status
from rdflib import Literal, URIRef
from rdflib.term import Node

from buildingmotif.api.serializers.template import serialize
from buildingmotif.database.errors import ModelNotFound, TemplateNotFound
from buildingmotif.dataclasses import Model, Template
from buildingmotif.ingresses import CSVIngress, TemplateIngress

blueprint = Blueprint("templates", __name__)


@blueprint.route("", methods=(["GET"]))
def get_all_templates() -> flask.Response:
    """Get all templates.

    :return: all templates
    :rtype: flask.Response
    """
    db_templates = current_app.building_motif.table_connection.get_all_db_templates()

    return jsonify(serialize(db_templates)), status.HTTP_200_OK


@blueprint.route("/<templates_id>", methods=(["GET"]))
def get_template(templates_id: int) -> flask.Response:
    """Get template by id.

    :param templates_id: template id
    :type templates_id: int
    :return: requested template
    :rtype: flask.Response
    """
    include_parameters = request.args.get("parameters", False)

    try:
        template = current_app.building_motif.table_connection.get_db_template(
            templates_id
        )
    except TemplateNotFound:
        return {"message": f"ID: {templates_id}"}, status.HTTP_404_NOT_FOUND

    return jsonify(serialize(template, include_parameters)), status.HTTP_200_OK


@blueprint.route("/<template_id>/evaluate/ingress", methods=(["POST"]))
def evaluate_ingress(template_id: int) -> flask.Response:
    # get template
    try:
        template = Template.load(template_id)
    except TemplateNotFound:
        return {"message": f"ID: {template_id}"}, status.HTTP_404_NOT_FOUND

    # get model
    model_id = request.args.get("model_id")
    if model_id is None:
        return {
            "message": "must contain query param 'model_id'"
        }, status.HTTP_400_BAD_REQUEST
    try:
        model = Model.load(model_id)
    except ModelNotFound:
        return {"message": f"ID: {model_id}"}, status.HTTP_404_NOT_FOUND

    # get file
    raw_data = flask.request.get_data()
    if raw_data is None:
        return {"message": "no file recieved."}, status.HTTP_400_NOT_FOUND

    # evaluate template
    try:
        data = StringIO(raw_data.decode("utf-8"))
        csv_ingress = CSVIngress(data=data)
        template_ingress = TemplateIngress(
            template.inline_dependencies(), None, csv_ingress
        )
        graph_or_template = template_ingress.graph(model.name)
    except Exception:
        return {"message": "Invalid csv."}, status.HTTP_400_BAD_REQUEST

    # parse bindings from input JSON
    if isinstance(graph_or_template, Template):
        graph = graph_or_template.body
    else:
        graph = graph_or_template

    return graph.serialize(format="ttl"), status.HTTP_200_OK


@blueprint.route("/<template_id>/evaluate/bindings", methods=(["POST"]))
def evaluate_bindings(template_id: int) -> flask.Response:
    """evaluate template with giving binding

    :param template_id: id of template
    :type template_id: int
    :return: evaluated Group
    :rtype: flask.Response
    """
    try:
        template = Template.load(template_id)
    except TemplateNotFound:
        return {"message": f"ID: {template_id}"}, status.HTTP_404_NOT_FOUND

    if request.content_type != "application/json":
        return {
            "message": "request content type must be json"
        }, status.HTTP_400_BAD_REQUEST

    model_id = request.get_json().get("model_id")
    if model_id is None:
        return {"message": "body must contain 'model_id'"}, status.HTTP_400_BAD_REQUEST
    try:
        model = Model.load(model_id)
    except ModelNotFound:
        return {"message": f"ID: {model_id}"}, status.HTTP_404_NOT_FOUND

    bindings = request.get_json().get("bindings")
    if bindings is None:
        return {"message": "body must contain 'bindings'"}, status.HTTP_400_BAD_REQUEST
    bindings = get_bindings(bindings)
    bindings = {k: model.name.rstrip("/") + "/" + v for k, v in bindings.items()}

    # parse bindings from input JSON
    graph_or_template = template.evaluate(bindings=bindings)
    if isinstance(graph_or_template, Template):
        graph = graph_or_template.body
    else:
        graph = graph_or_template

    return graph.serialize(format="ttl"), status.HTTP_200_OK


@blueprint.route("/<template_id>/body", methods=(["GET"]))
def get_template_body(template_id: int) -> flask.Response:
    """Get template body.

    :param template_id: template id
    :type template_id: int
    :return: template body
    :rtype: flask.Response
    """
    try:
        template: Template = Template.load(template_id)
    except TemplateNotFound:
        return {"message": f"ID: {template_id}"}, status.HTTP_404_NOT_FOUND

    return template.body.serialize(format="ttl"), status.HTTP_200_OK


def get_bindings(binding_dict) -> Dict[str, Node]:
    """type binding_dict values to nodes

    given:
        {name: {@id or @literal: value}}

    return:
        {name: typed value}

    :param binding_dict: untyped bindings
    :type binding_dict: dict
    :return: typed dict
    :rtype: Dict[str, Node]
    """
    bindings = {}
    for param, definition in binding_dict.items():
        if "@id" in definition:
            bindings[param] = URIRef(definition["@id"])
        if "@literal" in definition:
            dtype = definition.get("@datatype")
            bindings[param] = Literal(
                definition["@literal"], datatype=URIRef(dtype) if dtype else None
            )
    return bindings



from typing import List, Optional, Union

from typing_extensions import TypedDict

from buildingmotif.database.tables import DBLibrary, DBTemplate

LibraryDict = TypedDict(
    "LibraryDict",
    {
        "id": int,
        "name": str,
        "template_ids": List[int],
        "shape_collection_id": Optional[int],
    },
)


def serialize(
    param: Union[DBLibrary, List[DBLibrary]]
) -> Union[LibraryDict, List[LibraryDict]]:
    """Serialize one or more libraries into a TypedDict.

    :param param: one library or a list of libraries
    :type param: Union[DBLibrary, List[DBLibrary]]
    :raises ValueError: if invalid input
    :return: one JSON per serialized library
    :rtype: Union[DBLibrary, List[DBLibrary]]
    """
    if isinstance(param, DBLibrary):
        return _serialize(param)

    elif isinstance(param, list):
        return [_serialize(x) for x in param]

    raise ValueError("invalid input. Must be a DBLibrary or list of DBLibraries")


def _serialize(library: DBLibrary) -> LibraryDict:
    """Serialize a library into a TypedDict.

    :param library: library
    :type library: DBLibrary
    :return: serialized library
    :rtype: LibraryDict
    """
    templates: List[DBTemplate] = library.templates  # apleasing the type checker.

    return {
        "id": library.id,
        "name": library.name,
        "template_ids": [t.id for t in templates],
        "shape_collection_id": library.shape_collection_id,
    }



from typing import Callable, Dict

import rdflib

from buildingmotif.namespaces import BACNET, BRICK, PARAM, RDF, REF
from buildingmotif.utils import new_temporary_graph

RULES: Dict[str, Callable] = {}

"""
Notes:
- maybe more useful for authoring than for distribution
- "compile" into template definitions that then get distributed
- standardize the "subject" of the template: 'name' param
- have rules generate dependencies rather than introducing additional subjects?
- give the rules/macros access to the library and/or the template they
are associated with?:
  - new signature: def foo(library, template, *args, **kwargs) -> rdflib.Graph
- don't hardcode template names into these rules/macros --- pass in from user
"""


def rule(name: str):
    """This decorator is applied to a function in order to register that
    function as a rule that can be executed during template compilation.

    :param name: name
    :type name: str
    """

    def wrapper(func: Callable):
        RULES[name] = func
        return func

    return wrapper


@rule("body")
def body(*args, **kwargs) -> rdflib.Graph:
    """Returns the first value of the `body` argument as the template body.

    :return: graph of first value of `body` argument
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    G.parse(data=args[0], format="turtle")
    return G


@rule("type")
def hasType(*args, **kwargs) -> rdflib.Graph:
    """_summary_

    :return: _description_
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    for arg in args:
        arg = rdflib.URIRef(arg)
        G.add((PARAM["name"], RDF["type"], arg))
    return G


@rule("hasPoint")
def points(*args, **kwargs) -> rdflib.Graph:
    """_summary_

    :return: _description_
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    for param, pointtype in kwargs.items():
        pointtype = rdflib.URIRef(pointtype)
        G.add(
            (PARAM[param], RDF["type"], pointtype)
        )  # should really be a dependency on that type
        G.add((PARAM["name"], BRICK.hasPoint, PARAM[param]))
    return G


@rule("hasPart")
def parts(*args, **kwargs) -> rdflib.Graph:
    """_summary_

    :return: _description_
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    for param, pointtype in kwargs.items():
        pointtype = rdflib.URIRef(pointtype)
        G.add(
            (PARAM[param], RDF["type"], pointtype)
        )  # should really be a dependency on that type
        G.add((PARAM["name"], BRICK.hasPart, PARAM[param]))
    return G


@rule("upstream")
def isFedBy(*args, **kwargs) -> rdflib.Graph:
    """_summary_

    :return: _description_
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    assert kwargs is not None
    for param, equiptype in kwargs.items():
        equiptype = rdflib.URIRef(equiptype)
        G.add((PARAM[param], RDF["type"], equiptype))
        G.add((PARAM["name"], BRICK.isFedBy, PARAM[param]))
    return G


@rule("downstream")
def feeds(*args, **kwargs) -> rdflib.Graph:
    """_summary_

    :return: _description_
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    assert kwargs is not None
    for param, equiptype in kwargs.items():
        equiptype = rdflib.URIRef(equiptype)
        G.add((PARAM[param], RDF["type"], equiptype))
        G.add((PARAM["name"], BRICK.feeds, PARAM[param]))
    return G


@rule("bacnet")
def bacnet(*args, **kwargs) -> rdflib.Graph:
    """Adds template fields for each of the param names in `args` to ensure
    they have BACnet references.

    :return: a graph with BACnet references added
    :rtype: rdflib.Graph
    """
    G = new_temporary_graph()
    G.add((PARAM["BACnet_device"], RDF["type"], BACNET.BACnetDevice))
    for param in args:
        ref_name = rdflib.BNode()
        object_name = PARAM[param + "_object"]
        G.add((PARAM[param], REF.hasExternalReference, ref_name))
        G.add((ref_name, BACNET["object-identifier"], object_name))
        G.add((ref_name, BACNET["objectOf"], PARAM["BACnet_device"]))
    return G


def compile_template_spec(spec: Dict) -> Dict:
    """Compiles a template specification into the body of a template by
    applying rules defined in this file and others.

    :param spec: the Python dictionary that is the result of decoding the
        YAML structure (usually a dictionary or list) defining the
        template
    :type spec: Dict
    :raises ValueError: if unknown rule
    :return: the template spec containing the compiled body
    :rtype: Dict
    """
    # extract metadata that doesn't correspond to rule executions
    deps = spec.pop("dependencies", [])
    optionals = spec.pop("optional", [])

    # compile the template's body from the rules
    body = new_temporary_graph()
    existing_body = spec.pop("body", None)
    if existing_body is not None:
        body.parse(data=existing_body, format="turtle")
    for rule_name in tuple(spec.keys()):
        rule_args = spec.pop(rule_name)
        if rule_name not in RULES:
            raise ValueError(f"Unknown rule {rule_name}")
        if isinstance(rule_args, dict):
            G = RULES[rule_name](**rule_args)
        elif isinstance(rule_args, list):
            G = RULES[rule_name](*rule_args)
        else:
            G = RULES[rule_name](*[rule_args])
        body += G

    # put the metadata back
    spec["dependencies"] = deps
    spec["body"] = body
    spec["optional"] = optionals
    return spec
